"""Implementation of Rule L003."""
from typing import List, Optional, Sequence, Tuple

from sqlfluff.core.parser import WhitespaceSegment
from sqlfluff.core.parser.segments import BaseSegment, RawSegment
from sqlfluff.core.rules.base import BaseRule, LintResult, LintFix, RuleContext
from sqlfluff.core.rules.doc_decorators import (
    document_fix_compatible,
    document_configuration,
)
from sqlfluff.core.templaters import TemplatedFile


@document_fix_compatible
@document_configuration
class Rule_L003(BaseRule):
    """Indentation not consistent with previous lines.

    Note:
        This rule used to be _"Indentation length is not a multiple
        of `tab_space_size`"_, but was changed to be much smarter.

    | **Anti-pattern**
    | The • character represents a space.
    | In this example, the third line contains five spaces instead of four.

    .. code-block:: sql
       :force:

        SELECT
        ••••a,
        •••••b
        FROM foo


    | **Best practice**
    | Change the indentation to use a multiple of four spaces.

    .. code-block:: sql
       :force:

        SELECT
        ••••a,
        ••••b
        FROM foo

    """

    _works_on_unparsable = False
    _ignore_types: List[str] = ["script_content"]
    config_keywords = ["tab_space_size", "indent_unit"]

    @staticmethod
    def _make_indent(
        num: int = 1, tab_space_size: int = 4, indent_unit: str = "space"
    ) -> str:
        if indent_unit == "tab":
            base_unit = "\t"
        elif indent_unit == "space":
            base_unit = " " * tab_space_size
        else:
            raise ValueError(
                f"Parameter indent_unit has unexpected value: `{indent_unit}`. Expected `tab` or `space`."
            )
        return base_unit * num

    @staticmethod
    def _indent_size(segments: Sequence[RawSegment], tab_space_size: int = 4) -> int:
        indent_size = 0
        for elem in segments:
            raw = elem.raw
            # convert to spaces for convenience (and hanging indents)
            raw = raw.replace("\t", " " * tab_space_size)
            indent_size += len(raw)
        return indent_size

    @classmethod
    def _reorder_raw_stack(
        cls,
        raw_stack: Tuple[RawSegment, ...],
        templated_file: Optional[TemplatedFile],
    ) -> Tuple[RawSegment, ...]:
        """Reorder raw_stack to simplify indentation logic.

        Context: The indentation logic was mostly designed to work with normal
        segment types. Templating introduces additional segments into the parse
        tree, often in the "wrong" place with respect to the indentation logic,
        for example, where do indent/dedent segments appear with respect to the
        segments that trigger indent/dedent behavior? This function reorders
        nodes locally (i.e. only within L003) to get the desired behavior.
        """

        def segment_info(idx: int) -> Tuple[str, Optional[str]]:
            """Helper function for sort_current_line()."""
            seg = current_line[idx]
            return seg.type, cls._get_element_template_info(seg, templated_file)

        def move_indent_before_templated() -> None:
            """Swap position of template and indent segment if code follows.

            This allows for correct indentation of templated table names in
            "FROM", for example:

            SELECT brand
            FROM
                {{ product }}

            """
            for idx in range(2, len(current_line)):
                if (
                    segment_info(idx - 2)
                    == (
                        "placeholder",
                        "templated",
                    )
                    and segment_info(idx - 1) == ("indent", None)
                    and segment_info(idx) == ("raw", None)
                ):
                    current_line[idx - 2], current_line[idx - 1] = (
                        current_line[idx - 1],
                        current_line[idx - 2],
                    )

        # Break raw_stack into lines.
        lines = []
        current_line = []
        for elem in raw_stack:
            if not elem.is_type("newline"):
                current_line.append(elem)
            else:
                move_indent_before_templated()
                current_line.append(elem)
                lines.append(current_line)
                current_line = []
        if current_line:
            move_indent_before_templated()
            lines.append(current_line)
        new_raw_stack = [s for line in lines for s in line]
        return tuple(new_raw_stack)

    @classmethod
    def _process_raw_stack(
        cls,
        raw_stack: Tuple[RawSegment, ...],
        memory: dict = None,
        tab_space_size: int = 4,
        templated_file: Optional[TemplatedFile] = None,
    ) -> dict:
        """Take the raw stack, split into lines and evaluate some stats."""
        raw_stack = cls._reorder_raw_stack(raw_stack, templated_file)
        indent_balance = 0
        line_no = 1
        in_indent = True
        indent_buffer: List[RawSegment] = []
        line_buffer: List[RawSegment] = []
        result_buffer = {}
        indent_size = 0
        line_indent_stack: List[int] = []
        this_indent_balance = 0
        clean_indent = False
        hanger_pos = None

        for elem in raw_stack:
            line_buffer.append(elem)
            # Pin indent_balance to above zero
            if indent_balance < 0:
                indent_balance = 0

            if elem.is_type("newline"):
                result_buffer[line_no] = {
                    "line_no": line_no,
                    # Using slicing to copy line_buffer here to be py2 compliant
                    "line_buffer": line_buffer[:],
                    "indent_buffer": indent_buffer,
                    "indent_size": indent_size,
                    # Indent balance is the indent at the start of the first content
                    "indent_balance": this_indent_balance,
                    "hanging_indent": hanger_pos if line_indent_stack else None,
                    # Clean indent is true if the line *ends* with an indent
                    # or has an indent in the initial whitespace.
                    "clean_indent": clean_indent,
                }
                line_no += 1
                indent_buffer = []
                line_buffer = []
                indent_size = 0
                in_indent = True
                line_indent_stack = []
                hanger_pos = None
                # Assume an unclean indent, but if the last line
                # ended with an indent then we might be ok.
                clean_indent = False
                # Was there an indent after the last code element of the previous line?
                for search_elem in reversed(result_buffer[line_no - 1]["line_buffer"]):  # type: ignore
                    if not search_elem.is_code and not search_elem.is_meta:
                        continue
                    elif search_elem.is_meta and search_elem.indent_val > 0:
                        clean_indent = True
                    break
            elif in_indent:
                if elem.is_type("whitespace"):
                    indent_buffer.append(elem)
                elif elem.is_meta and elem.indent_val != 0:  # type: ignore
                    indent_balance += elem.indent_val  # type: ignore
                    if elem.indent_val > 0:  # type: ignore
                        # a "clean" indent is one where it contains
                        # an increase in indentation? Can't quite
                        # remember the logic here. Let's go with that.
                        clean_indent = True
                else:
                    in_indent = False
                    this_indent_balance = indent_balance
                    indent_size = cls._indent_size(
                        indent_buffer, tab_space_size=tab_space_size
                    )
            elif elem.is_meta and elem.indent_val != 0:  # type: ignore
                indent_balance += elem.indent_val  # type: ignore
                if elem.indent_val > 0:  # type: ignore
                    # Keep track of the indent at the last ... indent
                    line_indent_stack.append(
                        cls._indent_size(line_buffer, tab_space_size=tab_space_size)
                    )
                    hanger_pos = None
                else:
                    # this is a dedent, we could still have a hanging indent,
                    # but only if there's enough on the stack
                    if line_indent_stack:
                        line_indent_stack.pop()
            elif elem.is_code:
                if hanger_pos is None:
                    hanger_pos = cls._indent_size(
                        line_buffer[:-1], tab_space_size=tab_space_size
                    )

            # If we hit the trigger element, stop processing.
            if memory and elem is memory["trigger"]:
                break

        # If we get to the end, and still have a buffer, add it on
        if line_buffer:
            result_buffer[line_no] = {
                "line_no": line_no,
                "line_buffer": line_buffer,
                "indent_buffer": indent_buffer,
                "indent_size": indent_size,
                "indent_balance": this_indent_balance,
                "hanging_indent": line_indent_stack.pop()
                if line_indent_stack
                else None,
                "clean_indent": clean_indent,
            }
        return result_buffer

    def _coerce_indent_to(
        self,
        desired_indent: str,
        current_indent_buffer: Tuple[RawSegment, ...],
        current_anchor: BaseSegment,
    ) -> List[LintFix]:
        """Generate fixes to make an indent a certain size."""
        # If there shouldn't be an indent at all, just delete.
        if len(desired_indent) == 0:
            fixes = [LintFix("delete", elem) for elem in current_indent_buffer]
        # If we don't have any indent and we should, then add a single
        elif len("".join(elem.raw for elem in current_indent_buffer)) == 0:
            fixes = [
                LintFix(
                    "create",
                    current_anchor,
                    WhitespaceSegment(
                        raw=desired_indent,
                    ),
                )
            ]
        # Otherwise edit the first element to be the right size
        else:
            # Edit the first element of this line's indent.
            fixes = [
                LintFix(
                    "edit",
                    current_indent_buffer[0],
                    WhitespaceSegment(
                        raw=desired_indent,
                    ),
                )
            ]
        return fixes

    @staticmethod
    def _strip_buffers(line_dict: dict) -> dict:
        """Strip a line dict of buffers for logging."""
        return {
            key: line_dict[key]
            for key in line_dict
            if key not in ("line_buffer", "indent_buffer")
        }

    @classmethod
    def _is_last_segment(
        cls,
        segment: BaseSegment,
        memory: dict,
        parent_stack: Tuple[BaseSegment, ...],
        siblings_post: Tuple[BaseSegment, ...],
    ) -> bool:
        """Returns True if 'segment' is the very last node in the parse tree."""
        if siblings_post:
            # We have subsequent siblings. Not finished.
            return False
        elif parent_stack:
            # No subsequent siblings. Our parent is finished.
            memory["finished"].add(parent_stack[-1])
        if segment.segments:
            # We have children. Not finished.
            return False

        # We have no subsequent siblings or children. If all our parents are
        # finished, the whole parse tree is finished.
        for parent in parent_stack:
            if parent not in memory["finished"]:
                return False
        return True

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Indentation not consistent with previous lines.

        To set the default tab size, set the `tab_space_size` value
        in the appropriate configuration.

        We compare each line (first non-whitespace element of the
        line), with the indentation of previous lines. The presence
        (or lack) of indent or dedent meta-characters indicate whether
        the indent is appropriate.

        - Any line is assessed by the indent level at the first non
          whitespace element.
        - Any increase in indentation may be _up to_ the number of
          indent characters.
        - Any line must be in line with the previous line which had
          the same indent balance at its start.
        - Apart from "whole" indents, a "hanging" indent is possible
          if the line starts in line with either the indent of the
          previous line or if it starts at the same indent as the *last*
          indent meta segment in the previous line.

        """
        # Config type hints
        self.tab_space_size: int
        self.indent_unit: str

        raw_stack = context.raw_stack

        # We ignore certain types (e.g. non-SQL scripts in functions)
        # so check if on ignore list
        if context.segment.type in self._ignore_types:
            return LintResult()
        for parent in context.parent_stack:
            if parent.type in self._ignore_types:
                return LintResult()

        # Memory keeps track of what we've seen
        if not context.memory:
            memory: dict = {
                # in_indent keeps track of whether we're in an indent right now
                "in_indent": True,
                # problem_lines keeps track of lines with problems so that we
                # don't compare to them.
                "problem_lines": [],
                # hanging_lines keeps track of hanging lines so that we don't
                # compare to them when assessing indent.
                "hanging_lines": [],
                # comment_lines keeps track of lines which are all comment.
                "comment_lines": [],
                # segments we've seen the last child of
                "finished": set(),
                # First non-whitespace node on a line.
                "trigger": None,
            }
        else:
            memory = context.memory

        if context.segment.is_type("newline"):
            memory["in_indent"] = True
        elif memory["in_indent"]:
            if context.segment.is_type("whitespace"):
                # it's whitespace, carry on
                pass
            elif context.segment.segments or (context.segment.is_meta and context.segment.indent_val != 0):  # type: ignore
                # it's not a raw segment or placeholder. Carry on.
                pass
            else:
                memory["in_indent"] = False
                # we're found a non-whitespace element. This is our trigger,
                # which we'll handle after this if-statement
                memory["trigger"] = context.segment
        else:
            # Not in indent and not a newline, don't trigger here.
            pass

        # Is this the last segment? If so, need to "flush" any leftovers.
        is_last = self._is_last_segment(
            context.segment, memory, context.parent_stack, context.siblings_post
        )

        if not context.segment.is_type("newline") and not is_last:
            # We only process complete lines or on the very last segment
            # (since there may not be a newline on the very last line)..
            return LintResult(memory=memory)

        if raw_stack and raw_stack[-1] is not context.segment:
            raw_stack = raw_stack + (context.segment,)
        res = self._process_raw_stack(
            raw_stack,
            memory,
            tab_space_size=self.tab_space_size,
            templated_file=context.templated_file,
        )

        if res:
            # Saw a newline or end of parse tree. Is the current line empty?
            trigger_segment = memory["trigger"]
            if trigger_segment:
                # Not empty. Process it.
                result = self._process_current_line(res, memory)
                if context.segment.is_type("newline"):
                    memory["trigger"] = None
                return result
        return LintResult(memory=memory)

    def _process_current_line(self, res: dict, memory: dict) -> LintResult:
        """Checks indentation of one line of code, returning a LintResult.

        The _eval() function calls it for the current line of code:
        - When passed a newline segment (thus ending a line)
        - When passed the *final* segment in the entire parse tree (which may
          not be a newline)
        """
        this_line_no = max(res.keys())
        this_line = res.pop(this_line_no)
        self.logger.debug(
            "Evaluating line #%s. %s",
            this_line_no,
            # Don't log the line or indent buffer, it's too noisy.
            self._strip_buffers(this_line),
        )
        trigger_segment = memory["trigger"]

        # Is this line just comments? (Disregard trailing newline if present.)
        check_comment_line = this_line["line_buffer"]
        if check_comment_line and all(
            seg.is_type(
                "whitespace", "comment", "indent"  # dedent is a subtype of indent
            )
            for seg in check_comment_line
        ):
            # Comment line, deal with it later.
            memory["comment_lines"].append(this_line_no)
            self.logger.debug("    Comment Line. #%s", this_line_no)
            return LintResult(memory=memory)

        # Is it a hanging indent?
        # Find last meaningful line indent.
        last_code_line = None
        for k in sorted(res.keys(), reverse=True):
            if any(seg.is_code for seg in res[k]["line_buffer"]):
                last_code_line = k
                break

        if len(res) > 0 and last_code_line:
            last_line_hanger_indent = res[last_code_line]["hanging_indent"]
            # Let's just deal with hanging indents here.
            if (
                # NB: Hangers are only allowed if there was content after the last
                # indent on the previous line. Otherwise it's just an indent.
                this_line["indent_size"] == last_line_hanger_indent
                # Or they're if the indent balance is the same and the indent is the
                # same AND the previous line was a hanger
                or (
                    this_line["indent_size"] == res[last_code_line]["indent_size"]
                    and this_line["indent_balance"]
                    == res[last_code_line]["indent_balance"]
                    and last_code_line in memory["hanging_lines"]
                )
            ) and (
                # There MUST also be a non-zero indent. Otherwise we're just on the baseline.
                this_line["indent_size"]
                > 0
            ):
                # This is a HANGER
                memory["hanging_lines"].append(this_line_no)
                self.logger.debug("    Hanger Line. #%s", this_line_no)
                self.logger.debug(
                    "    Last Line: %s", self._strip_buffers(res[last_code_line])
                )
                return LintResult(memory=memory)

        # Is this an indented first line?
        elif len(res) == 0:
            if this_line["indent_size"] > 0:
                self.logger.debug("    Indented First Line. #%s", this_line_no)
                return LintResult(
                    anchor=trigger_segment,
                    memory=memory,
                    description="First line has unexpected indent",
                    fixes=[
                        LintFix("delete", elem) for elem in this_line["indent_buffer"]
                    ],
                )

        # Assuming it's not a hanger, let's compare it to the other previous
        # lines. We do it in reverse so that closer lines are more relevant.
        for k in sorted(res.keys(), reverse=True):

            # Is this a problem line?
            if k in memory["problem_lines"] + memory["hanging_lines"]:
                # Skip it if it is
                continue

            # Is this an empty line?
            if not any(elem.is_code for elem in res[k]["line_buffer"]):
                # Skip if it is
                continue

            # Work out the difference in indent
            indent_diff = this_line["indent_balance"] - res[k]["indent_balance"]
            # If we're comparing to a previous, more deeply indented line, then skip and keep looking.
            if indent_diff < 0:
                continue
            # Is the indent balance the same?
            elif indent_diff == 0:
                self.logger.debug("    [same indent balance] Comparing to #%s", k)
                if this_line["indent_size"] != res[k]["indent_size"]:
                    # Indents don't match even though balance is the same...
                    memory["problem_lines"].append(this_line_no)

                    # Work out desired indent
                    if res[k]["indent_size"] == 0:
                        desired_indent = ""
                    elif this_line["indent_size"] == 0:
                        desired_indent = self._make_indent(
                            indent_unit=self.indent_unit,
                            tab_space_size=self.tab_space_size,
                        )
                    else:
                        # The previous indent.
                        desired_indent = "".join(
                            elem.raw for elem in res[k]["indent_buffer"]
                        )

                    # Make fixes
                    fixes = self._coerce_indent_to(
                        desired_indent=desired_indent,
                        current_indent_buffer=this_line["indent_buffer"],
                        current_anchor=trigger_segment,
                    )
                    self.logger.debug(
                        "    !! Indentation does not match #%s. Fixes: %s", k, fixes
                    )
                    return LintResult(
                        anchor=trigger_segment,
                        memory=memory,
                        description="Indentation not consistent with line #{}".format(
                            k
                        ),
                        # See above for logic
                        fixes=fixes,
                    )
            # Are we at a deeper indent?
            elif indent_diff > 0:
                self.logger.debug("    [deeper indent balance] Comparing to #%s", k)
                # NB: We shouldn't need to deal with correct hanging indents
                # here, they should already have been dealt with before. We
                # may still need to deal with *creating* hanging indents if
                # appropriate.
                self.logger.debug(
                    "    Comparison Line: %s", self._strip_buffers(res[k])
                )

                # Check to see if we've got a whole number of multiples. If
                # we do then record the number for later, otherwise raise
                # an error. We do the comparison here so we have a reference
                # point to do the repairs. We need a sensible previous line
                # to base the repairs off. If there's no indent at all, then
                # we should also take this route because there SHOULD be one.
                if this_line["indent_size"] % self.tab_space_size != 0:
                    memory["problem_lines"].append(this_line_no)

                    # The default indent is the one just reconstructs it from
                    # the indent size.
                    default_indent = "".join(
                        elem.raw for elem in res[k]["indent_buffer"]
                    ) + self._make_indent(
                        indent_unit=self.indent_unit,
                        tab_space_size=self.tab_space_size,
                        num=indent_diff,
                    )
                    # If we have a clean indent, we can just add steps in line
                    # with the difference in the indent buffers. simples.
                    if this_line["clean_indent"]:
                        self.logger.debug("        Use clean indent.")
                        desired_indent = default_indent
                    # If we have the option of a hanging indent then use it.
                    elif res[k]["hanging_indent"]:
                        self.logger.debug("        Use hanging indent.")
                        desired_indent = " " * res[k]["hanging_indent"]
                    else:  # pragma: no cover
                        self.logger.debug("        Use default indent.")
                        desired_indent = default_indent

                    # Make fixes
                    fixes = self._coerce_indent_to(
                        desired_indent=desired_indent,
                        current_indent_buffer=this_line["indent_buffer"],
                        current_anchor=trigger_segment,
                    )

                    return LintResult(
                        anchor=trigger_segment,
                        memory=memory,
                        description=(
                            "Indentation not hanging or a multiple of {} spaces"
                        ).format(self.tab_space_size),
                        fixes=fixes,
                    )
                else:
                    # We'll need this value later.
                    this_indent_num = this_line["indent_size"] // self.tab_space_size

                # We know that the indent balance is higher, what actually is
                # the difference in indent counts? It should be a whole number
                # if we're still here.
                comp_indent_num = res[k]["indent_size"] // self.tab_space_size

                # The indent number should be at least 1, and can be UP TO
                # and including the difference in the indent balance.
                if comp_indent_num == this_indent_num:
                    # We have two lines indented the same, but with a different starting
                    # indent balance. This is either a problem OR a sign that one of the
                    # opening indents wasn't used. We account for the latter and then
                    # have a violation if that wasn't the case.

                    # Does the comparison line have enough unused indent to get us back
                    # to where we need to be? NB: This should only be applied if this is
                    # a CLOSING bracket.

                    # First work out if we have some closing brackets, and if so, how many.
                    b_idx = 0
                    b_num = 0
                    while True:
                        if len(this_line["line_buffer"][b_idx:]) == 0:
                            break

                        elem = this_line["line_buffer"][b_idx]
                        if not elem.is_code:
                            b_idx += 1
                            continue
                        else:
                            if elem.is_type("end_bracket", "end_square_bracket"):
                                b_idx += 1
                                b_num += 1
                                continue
                            break  # pragma: no cover

                    if b_num >= indent_diff:
                        # It does. This line is fine.
                        pass
                    else:
                        # It doesn't. That means we *should* have an indent when compared to
                        # this line and we DON'T.
                        memory["problem_lines"].append(this_line_no)
                        return LintResult(
                            anchor=trigger_segment,
                            memory=memory,
                            description="Indent expected and not found compared to line #{}".format(
                                k
                            ),
                            # Add in an extra bit of whitespace for the indent
                            fixes=[
                                LintFix(
                                    "create",
                                    trigger_segment,
                                    WhitespaceSegment(
                                        raw=self._make_indent(
                                            indent_unit=self.indent_unit,
                                            tab_space_size=self.tab_space_size,
                                        ),
                                    ),
                                )
                            ],
                        )
                elif this_indent_num < comp_indent_num:
                    memory["problem_lines"].append(this_line_no)
                    return LintResult(
                        anchor=trigger_segment,
                        memory=memory,
                        description="Line under-indented compared to line #{}".format(
                            k
                        ),
                        fixes=[
                            LintFix(
                                "create",
                                trigger_segment,
                                WhitespaceSegment(
                                    # Make the minimum indent for it to be ok.
                                    raw=self._make_indent(
                                        num=comp_indent_num - this_indent_num,
                                        indent_unit=self.indent_unit,
                                        tab_space_size=self.tab_space_size,
                                    ),
                                ),
                            )
                        ],
                    )
                elif this_indent_num > comp_indent_num + indent_diff:
                    # Calculate the lowest ok indent:
                    desired_indent = self._make_indent(
                        num=comp_indent_num - this_indent_num,
                        indent_unit=self.indent_unit,
                        tab_space_size=self.tab_space_size,
                    )

                    # Make fixes
                    fixes = self._coerce_indent_to(
                        desired_indent=desired_indent,
                        current_indent_buffer=this_line["indent_buffer"],
                        current_anchor=trigger_segment,
                    )

                    memory["problem_lines"].append(this_line_no)
                    return LintResult(
                        anchor=trigger_segment,
                        memory=memory,
                        description="Line over-indented compared to line #{}".format(k),
                        fixes=fixes,
                    )

            # This was a valid comparison, so if it doesn't flag then
            # we can assume that we're ok.
            self.logger.debug("    Indent deemed ok comparing to #%s", k)

            # Given that this line is ok, consider if the preceding lines are
            # comments. If they are, lint the indentation of the comment(s).
            fixes = []
            for n in range(this_line_no - 1, -1, -1):
                if n in memory["comment_lines"]:
                    # The previous line WAS a comment.
                    prev_line = res[n]
                    if this_line["indent_size"] != prev_line["indent_size"]:
                        # It's not aligned.
                        # Find the anchor first.
                        anchor: BaseSegment = None  # type: ignore
                        for seg in prev_line["line_buffer"]:
                            if seg.is_type("comment"):
                                anchor = seg
                                break
                        # Make fixes.
                        fixes += self._coerce_indent_to(
                            desired_indent="".join(
                                elem.raw for elem in this_line["indent_buffer"]
                            ),
                            current_indent_buffer=prev_line["indent_buffer"],
                            current_anchor=anchor,
                        )

                        memory["problem_lines"].append(n)
                else:
                    break

            if fixes:
                return LintResult(
                    anchor=anchor,
                    memory=memory,
                    description="Comment not aligned with following line.",
                    fixes=fixes,
                )

            # Otherwise all good.
            return LintResult(memory=memory)

            # NB: At shallower indents, we don't check, we just check the
            # previous lines with the same balance. Deeper indents can check
            # themselves.

        # If we get to here, then we're all good for now.
        return LintResult(memory=memory)

    @classmethod
    def _get_element_template_info(
        cls, elem: BaseSegment, templated_file: Optional[TemplatedFile]
    ) -> Optional[str]:
        if elem.is_type("placeholder"):
            if templated_file is None:
                raise ValueError("Parameter templated_file cannot be: None.")
            slices = templated_file.raw_slices_spanning_source_slice(
                elem.pos_marker.source_slice
            )
            if slices:
                return slices[0].slice_type
        return None



import pytest
import pandas

from pvlib.tests import conftest
from pvlib.tests.conftest import fail_on_pvlib_version

from pvlib._deprecation import pvlibDeprecationWarning, deprecated

@pytest.mark.xfail(strict=True,
                   reason='fail_on_pvlib_version should cause test to fail')
@fail_on_pvlib_version('0.0')
def test_fail_on_pvlib_version():
    pass


@fail_on_pvlib_version('100000.0')
def test_fail_on_pvlib_version_pass():
    pass


@pytest.mark.xfail(strict=True, reason='ensure that the test is called')
@fail_on_pvlib_version('100000.0')
def test_fail_on_pvlib_version_fail_in_test():
    raise Exception


# set up to test using fixtures with function decorated with
# conftest.fail_on_pvlib_version
@pytest.fixture()
def some_data():
    return "some data"


def alt_func(*args):
    return args


deprec_func = deprecated('350.8', alternative='alt_func',
                         name='deprec_func', removal='350.9')(alt_func)


@fail_on_pvlib_version('350.9')
def test_use_fixture_with_decorator(some_data):
    # test that the correct data is returned by the some_data fixture
    assert some_data == "some data"
    with pytest.warns(pvlibDeprecationWarning):  # test for deprecation warning
        deprec_func(some_data)


@pytest.mark.parametrize('function_name', ['assert_index_equal',
                                           'assert_series_equal',
                                           'assert_frame_equal'])
@pytest.mark.parametrize('pd_version', ['1.0.0', '1.1.0'])
@pytest.mark.parametrize('check_less_precise', [True, False])
def test__check_pandas_assert_kwargs(mocker, function_name, pd_version,
                                     check_less_precise):
    # test that conftest._check_pandas_assert_kwargs returns appropriate
    # kwargs for the assert_x_equal functions

    # NOTE: be careful about mixing mocker.patch and pytest.MonkeyPatch!
    # they do not coordinate their cleanups, so it is safest to only
    # use one or the other.  GH #1447

    # patch the pandas assert; not interested in actually calling them,
    # plus we want to spy on how they get called.
    spy = mocker.patch('pandas.testing.' + function_name)
    # patch pd.__version__ to exercise the two branches in
    # conftest._check_pandas_assert_kwargs
    mocker.patch('pandas.__version__', new=pd_version)

    # finally, run the function and check what args got passed to pandas:
    assert_function = getattr(conftest, function_name)
    args = [None, None]
    assert_function(*args, check_less_precise=check_less_precise)
    if pd_version == '1.1.0':
        tol = 1e-3 if check_less_precise else 1e-5
        expected_kwargs = {'atol': tol, 'rtol': tol}
    else:
        expected_kwargs = {'check_less_precise': check_less_precise}

    spy.assert_called_once_with(*args, **expected_kwargs)



import pandas as pd
import pytest
import pytz

from pvlib.iotools import midc
from conftest import DATA_DIR, RERUNS, RERUNS_DELAY


@pytest.fixture
def test_mapping():
    return {
        'Direct Normal [W/m^2]': 'dni',
        'Global PSP [W/m^2]': 'ghi',
        'Rel Humidity [%]': 'relative_humidity',
        'Temperature @ 2m [deg C]': 'temp_air',
        'Non Existant': 'variable',
    }


MIDC_TESTFILE = DATA_DIR / 'midc_20181014.txt'
MIDC_RAW_TESTFILE = DATA_DIR / 'midc_raw_20181018.txt'
MIDC_RAW_SHORT_HEADER_TESTFILE = (
    DATA_DIR / 'midc_raw_short_header_20191115.txt')

# TODO: not used, remove?
# midc_network_testfile = ('https://midcdmz.nrel.gov/apps/data_api.pl'
#                          '?site=UAT&begin=20181018&end=20181019')


def test_midc_format_index():
    data = pd.read_csv(MIDC_TESTFILE)
    data = midc.format_index(data)
    start = pd.Timestamp("20181014 00:00")
    start = start.tz_localize("MST")
    end = pd.Timestamp("20181014 23:59")
    end = end.tz_localize("MST")
    assert type(data.index) == pd.DatetimeIndex
    assert data.index[0] == start
    assert data.index[-1] == end


def test_midc_format_index_tz_conversion():
    data = pd.read_csv(MIDC_TESTFILE)
    data = data.rename(columns={'MST': 'PST'})
    data = midc.format_index(data)
    assert data.index[0].tz == pytz.timezone('Etc/GMT+8')


def test_midc_format_index_raw():
    data = pd.read_csv(MIDC_RAW_TESTFILE)
    data = midc.format_index_raw(data)
    start = pd.Timestamp('20181018 00:00')
    start = start.tz_localize('MST')
    end = pd.Timestamp('20181018 23:59')
    end = end.tz_localize('MST')
    assert data.index[0] == start
    assert data.index[-1] == end


def test_read_midc_var_mapping_as_arg(test_mapping):
    data = midc.read_midc(MIDC_TESTFILE, variable_map=test_mapping)
    assert 'ghi' in data.columns
    assert 'temp_air' in data.columns


@pytest.mark.remote_data
@pytest.mark.flaky(reruns=RERUNS, reruns_delay=RERUNS_DELAY)
def test_read_midc_raw_data_from_nrel():
    start_ts = pd.Timestamp('20181018')
    end_ts = pd.Timestamp('20181019')
    var_map = midc.MIDC_VARIABLE_MAP['UAT']
    data = midc.read_midc_raw_data_from_nrel('UAT', start_ts, end_ts, var_map)
    for k, v in var_map.items():
        assert v in data.columns
    assert data.index.size == 2880


def test_read_midc_header_length_mismatch(mocker):
    mock_data = mocker.MagicMock()
    with MIDC_RAW_SHORT_HEADER_TESTFILE.open() as f:
        mock_data.text = f.read()
    mocker.patch('pvlib.iotools.midc.requests.get',
                 return_value=mock_data)
    start = pd.Timestamp('2019-11-15T00:00:00-06:00')
    end = pd.Timestamp('2019-11-15T23:59:00-06:00')
    data = midc.read_midc_raw_data_from_nrel('', start, end)
    assert isinstance(data.index, pd.DatetimeIndex)
    assert data.index[0] == start
    assert data.index[-1] == end



"""
This module contains implementations of PV module and array electrical models.

These models are used to predict the electrical behavior of pv modules
or collections of pv modules (arrays).  The primary inputs are
effective irradiance and operating temperature and the outputs may range from
power or efficiency at the maximum power point to complete IV curves.
Supporting functions and parameter fitting functions may also be found here.
"""

import numpy as np
from scipy.optimize import curve_fit
from scipy.special import exp10


def pvefficiency_adr(effective_irradiance, temp_cell,
                     k_a, k_d, tc_d, k_rs, k_rsh):
    '''
    Calculate PV module efficiency using the ADR model.

    The efficiency varies with irradiance and operating temperature
    and is determined by 5 model parameters as described in [1]_.

    Parameters
    ----------
    effective_irradiance : numeric, non-negative
        The effective irradiance incident on the PV module. [W/m^2]

    temp_cell : numeric
        The PV module operating temperature. [°C]

    k_a : numeric
        Absolute scaling factor, which is equal to the efficiency at
        reference conditions. This factor allows the model to be used
        with relative or absolute efficiencies, and to accommodate data sets
        which are not perfectly normalized but have a slight bias at
        the reference conditions. [unitless]

    k_d : numeric, negative
        “Dark irradiance” or diode coefficient which influences the voltage
        increase with irradiance. [unitless]

    tc_d : numeric
        Temperature coefficient of the diode coefficient, which indirectly
        influences voltage. Because it is the only temperature coefficient
        in the model, its value will also reflect secondary temperature
        dependencies that are present in the PV module. [unitless]

    k_rs : numeric
        Series resistance loss coefficient. Because of the normalization
        it can be read as a power loss fraction at reference conditions.
        For example, if ``k_rs`` is 0.05, the internal loss assigned to the
        series resistance has a magnitude equal to 5% of the module output.
        [unitless]

    k_rsh : numeric
        Shunt resistance loss coefficient. Can be interpreted as a power
        loss fraction at reference conditions like ``k_rs``.
        [unitless]

    Returns
    -------
    eta : numeric
        The efficiency of the module at the specified irradiance and
        temperature.

    Notes
    -----
    Efficiency values ``eta`` may be absolute or relative, and may be expressed
    as percent or per unit.  This is determined by the efficiency data
    used to derive values for the 5 model parameters.  The first model
    parameter ``k_a`` is equal to the efficiency at STC and therefore
    indicates the efficiency scale being used. ``k_a`` can also be changed
    freely to adjust the scale, or to change the module to a slightly
    higher or lower efficiency class.

    All arguments may be scalars or array-like. If multiple arguments
    are array-like they must be the same shape or broadcastable to the
    same shape.

    See also
    --------
    pvlib.pvarray.fit_pvefficiency_adr

    References
    ----------
    .. [1] A. Driesse and J. S. Stein, "From IEC 61853 power measurements
       to PV system simulations", Sandia Report No. SAND2020-3877, 2020.
       :doi:`10.2172/1615179`

    .. [2] A. Driesse, M. Theristis and J. S. Stein, "A New Photovoltaic Module
       Efficiency Model for Energy Prediction and Rating," in IEEE Journal
       of Photovoltaics, vol. 11, no. 2, pp. 527-534, March 2021.
       :doi:`10.1109/JPHOTOV.2020.3045677`

    Examples
    --------
    >>> pvefficiency_adr([1000, 200], 25,
            k_a=100, k_d=-6.0, tc_d=0.02, k_rs=0.05, k_rsh=0.10)
    array([100.        ,  92.79729308])

    >>> pvefficiency_adr([1000, 200], 25,
            k_a=1.0, k_d=-6.0, tc_d=0.02, k_rs=0.05, k_rsh=0.10)
    array([1.        , 0.92797293])

    '''
    # Contributed by Anton Driesse (@adriesse), PV Performance Labs, Dec. 2022
    # Adapted from https://github.com/adriesse/pvpltools-python

    k_a = np.array(k_a)
    k_d = np.array(k_d)
    tc_d = np.array(tc_d)
    k_rs = np.array(k_rs)
    k_rsh = np.array(k_rsh)

    # normalize the irradiance
    G_REF = np.array(1000.)
    s = effective_irradiance / G_REF

    # obtain the difference from reference temperature
    T_REF = np.array(25.)
    dt = temp_cell - T_REF

    # equation 29 in JPV
    s_o     = exp10(k_d + (dt * tc_d))                             # noQA: E221
    s_o_ref = exp10(k_d)

    # equation 28 and 30 in JPV
    # the constant k_v does not appear here because it cancels out
    v  = np.log(s / s_o     + 1)                                   # noQA: E221
    v /= np.log(1 / s_o_ref + 1)

    # equation 25 in JPV
    eta = k_a * ((1 + k_rs + k_rsh) * v - k_rs * s - k_rsh * v**2)

    return eta


def fit_pvefficiency_adr(effective_irradiance, temp_cell, eta,
                         dict_output=True, **kwargs):
    """
    Determine the parameters of the ADR module efficiency model by non-linear
    least-squares fit to lab or field measurements.

    Parameters
    ----------
    effective_irradiance : numeric, non-negative
        Effective irradiance incident on the PV module. [W/m^2]

    temp_cell : numeric
        PV module operating temperature. [°C]

    eta : numeric
        Efficiency of the PV module at the specified irradiance and
        temperature(s). [unitless] or [%]

    dict_output : boolean, optional
        When True (default), return the result as a dictionary; when False,
        return the result as a numpy array.

    kwargs :
        Optional keyword arguments passed to `scipy.optimize.curve_fit`.
        These kwargs can over-ride some options set within this function,
        which could be interesting for very advanced users.

    Returns
    -------
    popt : array or dict
        Optimal values for the parameters.

    Notes
    -----
    The best fits are obtained when the lab or field data include a wide range
    of both irradiance and temperature values.  A minimal data set
    would consist of 6 operating points covering low, medium and high
    irradiance levels at two operating temperatures.

    See also
    --------
    pvlib.pvarray.pvefficiency_adr
    scipy.optimize.curve_fit

    """
    # Contributed by Anton Driesse (@adriesse), PV Performance Labs, Dec. 2022
    # Adapted from https://github.com/adriesse/pvpltools-python

    irradiance = np.asarray(effective_irradiance, dtype=float).reshape(-1)
    temperature = np.asarray(temp_cell, dtype=float).reshape(-1)
    eta = np.asarray(eta, dtype=float).reshape(-1)

    eta_max = np.max(eta)

    P_NAMES = ['k_a', 'k_d', 'tc_d', 'k_rs', 'k_rsh']
    P_MAX   = [+np.inf,   0, +0.1, 1, 1]                           # noQA: E221
    P_MIN   = [0,       -12, -0.1, 0, 0]                           # noQA: E221
    P0      = [eta_max,  -6,  0.0, 0, 0]                           # noQA: E221
    P_SCALE = [eta_max,  10,  0.1, 1, 1]

    SIGMA = 1 / np.sqrt(irradiance / 1000)

    fit_options = dict(p0=P0,
                       bounds=[P_MIN, P_MAX],
                       method='trf',
                       x_scale=P_SCALE,
                       loss='soft_l1',
                       f_scale=eta_max * 0.05,
                       sigma=SIGMA,
                       )

    fit_options.update(kwargs)

    def adr_wrapper(xdata, *params):
        return pvefficiency_adr(*xdata, *params)

    result = curve_fit(adr_wrapper,
                       xdata=[irradiance, temperature],
                       ydata=eta,
                       **fit_options,
                       )
    popt = result[0]

    if dict_output:
        return dict(zip(P_NAMES, popt))
    else:
        return popt



"""Individual segment parsers.

Matchable objects which return individual segments.
"""

import re
from typing import Type, Optional, List, Tuple, Union

from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.matchable import Matchable
from sqlfluff.core.parser.match_result import MatchResult
from sqlfluff.core.parser.segments import RawSegment, BaseSegment


class StringParser(Matchable):
    """An object which matches and returns raw segments based on strings."""

    # Meta segments are handled seperately. All StringParser elements
    # are assumed to be not meta.
    is_meta = False

    def __init__(
        self,
        template: str,
        raw_class: Type[RawSegment],
        name: Optional[str] = None,
        type: Optional[str] = None,
        optional: bool = False,
        **segment_kwargs,
    ):
        # String matchers are not case sensitive, so we make the template
        # uppercase on creation. If any SQL dialect is found to be case
        # sensitive for keywords, this could be extended to allow
        # case sensitivity.
        self.template = template.upper()
        self.raw_class = raw_class
        self.name = name
        self.type = type
        self.optional = optional
        self.segment_kwargs = segment_kwargs or {}

    def is_optional(self) -> bool:
        """Return whether this element is optional."""
        return self.optional

    def simple(self, parse_context: "ParseContext") -> Optional[List[str]]:
        """Return simple options for this matcher.

        Because string matchers are not case sensitive we can
        just return the template here.
        """
        return [self.template]

    def _is_first_match(self, segment: BaseSegment):
        """Does the segment provided match according to the current rules."""
        # Is the target a match and IS IT CODE.
        # The latter stops us accidentally matching comments.
        if self.template == segment.raw.upper() and segment.is_code:
            return True
        return False

    def _make_match_from_first_result(self, segments: Tuple[BaseSegment, ...]):
        """Make a MatchResult from the first segment in the given list.

        This is a helper function for reuse by other parsers.
        """
        # Construct the segment object
        new_seg = self.raw_class(
            raw=segments[0].raw,
            pos_marker=segments[0].pos_marker,
            type=self.type,
            name=self.name,
            **self.segment_kwargs,
        )
        # Return as a tuple
        return MatchResult((new_seg,), segments[1:])

    def match(
        self,
        segments: Union[BaseSegment, Tuple[BaseSegment, ...]],
        parse_context: "ParseContext",
    ) -> MatchResult:
        """Compare input segments for a match, return a `MatchResult`.

        Note: For matching here, we only consider the *first* element,
        because we assume that a keyword can only span one raw segment.
        """
        # If we've been passed the singular, make it a tuple
        if isinstance(segments, BaseSegment):
            segments = (segments,)

        # We're only going to match against the first element
        if len(segments) >= 1:
            # Is the first one already of this type?
            if (
                isinstance(segments[0], self.raw_class)
                and segments[0].name == self.name
                and segments[0].is_type(self.type)
            ):
                return MatchResult((segments[0],), segments[1:])
            # Does it match?
            elif self._is_first_match(segments[0]):
                return self._make_match_from_first_result(segments)
        return MatchResult.from_unmatched(segments)


class NamedParser(StringParser):
    """An object which matches and returns raw segments based on names."""

    def simple(cls, parse_context: ParseContext) -> Optional[List[str]]:
        """Does this matcher support a uppercase hash matching route?

        NamedParser segment does NOT for now. We might need to later for efficiency.

        There is a way that this *could* be enabled, by allowing *another*
        shortcut route, to look ahead at the names of upcoming segments,
        rather than their content.
        """
        return None

    def _is_first_match(self, segment: BaseSegment):
        """Does the segment provided match according to the current rules.

        NamedParser implements its own matching function where
        we assume that ._template is the `name` of a segment.
        """
        # Case sensitivity is not supported. Names are all
        # lowercase by convention.
        if self.template.lower() == segment.name.lower():
            return True
        return False


class RegexParser(StringParser):
    """An object which matches and returns raw segments based on a regex."""

    def __init__(
        self,
        template: str,
        raw_class: Type[RawSegment],
        name: Optional[str] = None,
        type: Optional[str] = None,
        optional: bool = False,
        anti_template: Optional[str] = None,
        **segment_kwargs,
    ):
        # Store the optional anti-template
        self.anti_template = anti_template
        super().__init__(
            template=template,
            raw_class=raw_class,
            name=name,
            type=type,
            optional=optional,
            **segment_kwargs,
        )

    def simple(cls, parse_context: ParseContext) -> Optional[List[str]]:
        """Does this matcher support a uppercase hash matching route?

        Regex segment does NOT for now. We might need to later for efficiency.
        """
        return None

    def _is_first_match(self, segment: BaseSegment):
        """Does the segment provided match according to the current rules.

        RegexParser implements its own matching function where
        we assume that ._template is a r"" string, and is formatted
        for use directly as a regex. This only matches on a single segment.
        """
        if len(segment.raw) == 0:  # pragma: no cover TODO?
            # If it's of zero length it's probably a meta segment.
            # In any case, it won't match here.
            return False
        # Try the regex. Case sensitivity is not supported.
        result = re.match(self.template, segment.raw_upper)
        if result:
            result_string = result.group(0)
            # Check that we've fully matched
            if result_string == segment.raw_upper:
                # Check that the anti_template (if set) hasn't also matched
                if self.anti_template and re.match(
                    self.anti_template, segment.raw_upper
                ):
                    return False
                else:
                    return True
        return False



"""
ASV benchmarks for infinite_sheds.py
"""

import numpy as np
import pandas as pd
from pvlib.bifacial import infinite_sheds
from pvlib import location, tracking


class InfiniteSheds:

    # benchmark variant parameters (run both vectorize=True and False)
    params = [True, False]
    param_names = ['vectorize']

    def setup(self, vectorize):
        self.times = pd.date_range(start='20180601', freq='1min',
                                   periods=1440)
        self.location = location.Location(40, -80)
        self.solar_position = self.location.get_solarposition(self.times)
        self.clearsky_irradiance = self.location.get_clearsky(
            self.times,
            solar_position=self.solar_position,
        )
        self.surface_tilt = 20
        self.surface_azimuth = 180
        self.gcr = 0.35
        self.height = 2.5
        self.pitch = 5.
        self.albedo = 0.2
        self.npoints = 100

        with np.errstate(invalid='ignore'):
            self.tracking = tracking.singleaxis(
                self.solar_position['apparent_zenith'],
                self.solar_position['azimuth'],
                axis_tilt=0,
                axis_azimuth=0,
                max_angle=60,
                backtrack=True,
                gcr=self.gcr
            )

    def time_get_irradiance_poa_fixed(self, vectorize):
        infinite_sheds.get_irradiance_poa(
            surface_tilt=self.surface_tilt,
            surface_azimuth=self.surface_azimuth,
            solar_zenith=self.solar_position['apparent_zenith'],
            solar_azimuth=self.solar_position['azimuth'],
            gcr=self.gcr,
            height=self.height,
            pitch=self.pitch,
            ghi=self.clearsky_irradiance['ghi'],
            dhi=self.clearsky_irradiance['dhi'],
            dni=self.clearsky_irradiance['dni'],
            albedo=self.albedo,
            npoints=self.npoints,
            vectorize=vectorize,
        )

    def time_get_irradiance_poa_tracking(self, vectorize):
        infinite_sheds.get_irradiance_poa(
            surface_tilt=self.tracking['surface_tilt'],
            surface_azimuth=self.tracking['surface_azimuth'],
            solar_zenith=self.solar_position['apparent_zenith'],
            solar_azimuth=self.solar_position['azimuth'],
            gcr=self.gcr,
            height=self.height,
            pitch=self.pitch,
            ghi=self.clearsky_irradiance['ghi'],
            dhi=self.clearsky_irradiance['dhi'],
            dni=self.clearsky_irradiance['dni'],
            albedo=self.albedo,
            npoints=self.npoints,
            vectorize=vectorize,
        )

    def time_get_irradiance_fixed(self, vectorize):
        infinite_sheds.get_irradiance(
            surface_tilt=self.surface_tilt,
            surface_azimuth=self.surface_azimuth,
            solar_zenith=self.solar_position['apparent_zenith'],
            solar_azimuth=self.solar_position['azimuth'],
            gcr=self.gcr,
            height=self.height,
            pitch=self.pitch,
            ghi=self.clearsky_irradiance['ghi'],
            dhi=self.clearsky_irradiance['dhi'],
            dni=self.clearsky_irradiance['dni'],
            albedo=self.albedo,
            npoints=self.npoints,
            vectorize=vectorize,
        )

    def time_get_irradiance_tracking(self, vectorize):
        infinite_sheds.get_irradiance(
            surface_tilt=self.tracking['surface_tilt'],
            surface_azimuth=self.tracking['surface_azimuth'],
            solar_zenith=self.solar_position['apparent_zenith'],
            solar_azimuth=self.solar_position['azimuth'],
            gcr=self.gcr,
            height=self.height,
            pitch=self.pitch,
            ghi=self.clearsky_irradiance['ghi'],
            dhi=self.clearsky_irradiance['dhi'],
            dni=self.clearsky_irradiance['dni'],
            albedo=self.albedo,
            npoints=self.npoints,
            vectorize=vectorize,
        )



#!/usr/bin/env python3
from __future__ import annotations

import json
import shutil
import statistics
import subprocess
import sys
import textwrap
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path

from docx import Document
from docx.shared import Pt

ROOT = Path(__file__).resolve().parents[1]
OUT_DIR = ROOT / "validation" / "fresh_effectiveness_20260326"
WORK_DIR = OUT_DIR / "workspaces"
REPORT_MD = OUT_DIR / "FRESH_TOOL_EFFECTIVENESS.md"
REPORT_JSON = OUT_DIR / "fresh_tool_effectiveness.json"
REPORT_DOCX = OUT_DIR / "WLBS_Fresh_Tool_Effectiveness_20260326.docx"

sys.path.insert(0, str(ROOT))
import wlbs_scan as ws  # noqa: E402


@dataclass
class Patch:
    relpath: str
    old: str
    new: str
    label: str


@dataclass
class Task:
    name: str
    split: str
    family: str
    symptom: str
    true_target: str
    project_files: dict[str, str]
    correct_patch: Patch
    wrong_symptom_patches: list[Patch]
    confidence_threshold: float
    expected_mode: str


def write(path: Path, content: str):
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(textwrap.dedent(content).lstrip("\n"), encoding="utf-8")


def apply_patch_text(root: Path, patch: Patch):
    path = root / patch.relpath
    text = path.read_text(encoding="utf-8")
    if patch.old not in text:
        raise ValueError(f"patch anchor not found in {path}: {patch.label}")
    path.write_text(text.replace(patch.old, patch.new, 1), encoding="utf-8")


def run_pytest(project_root: Path) -> dict:
    env = dict(**__import__("os").environ)
    env["PYTHONDONTWRITEBYTECODE"] = "1"
    proc = subprocess.run(
        [sys.executable, "-m", "pytest", "tests", "-q"],
        cwd=project_root,
        capture_output=True,
        text=True,
        encoding="utf-8",
        errors="replace",
        env=env,
        timeout=60,
    )
    output = (proc.stdout + "\n" + proc.stderr).strip()
    passed = failed = 0
    m_passed = __import__("re").search(r"(\d+)\s+passed", output)
    m_failed = __import__("re").search(r"(\d+)\s+failed", output)
    if m_passed:
        passed = int(m_passed.group(1))
    if m_failed:
        failed = int(m_failed.group(1))
    return {
        "ok": proc.returncode == 0,
        "returncode": proc.returncode,
        "passed": passed,
        "failed": failed,
        "summary": output,
    }


def run_cli(project_root: Path, args: list[str]) -> str:
    proc = subprocess.run(
        [sys.executable, str(ROOT / "wlbs_scan.py"), str(project_root), *args],
        capture_output=True,
        text=True,
        encoding="utf-8",
        errors="replace",
        timeout=60,
    )
    return (proc.stdout + ("\n" + proc.stderr if proc.stderr else "")).strip()


def fresh_store(root: Path, shared_memory: dict | None = None) -> ws.WorldLineStore:
    store = ws.WorldLineStore(root)
    if shared_memory:
        store.task_memory = json.loads(json.dumps(shared_memory.get("task_memory", {})))
        store.routing_policy = json.loads(json.dumps(shared_memory.get("routing_policy", {})))
        store._recompute_routing_stats()
    return store


def merge_memory(shared_memory: dict, store: ws.WorldLineStore):
    shared_memory["task_memory"] = json.loads(json.dumps(store.task_memory))
    shared_memory["routing_policy"] = json.loads(json.dumps(store.routing_policy))


def build_missing_key_task(
    *,
    name: str,
    split: str,
    family: str,
    root_mod: str,
    symptom_mod: str,
    missing_key: str,
    fallback_key: str,
    confidence_threshold: float,
) -> Task:
    root_file = f"{root_mod}.py"
    symptom_file = f"{symptom_mod}.py"
    tests_file = f"tests/test_{symptom_mod}.py"
    project_files = {
        root_file: f"""
        TABLE = {{
            "{fallback_key}": ["read", "write"]
        }}

        def get_rules(role):
            return TABLE[role]
        """,
        symptom_file: f"""
        import {root_mod}

        def can_delete(role="{missing_key}"):
            # WLBS_FRESH_SYMPTOM_BLOCK
            rules = {root_mod}.get_rules(role)
            return "delete" in rules

        def can_write(role="{missing_key}"):
            rules = {root_mod}.get_rules(role)
            return "write" in rules
        """,
        tests_file: f"""
        import {symptom_mod}

        def test_access_delete():
            assert {symptom_mod}.can_delete() is True

        def test_access_write():
            assert {symptom_mod}.can_write() is True
        """,
    }
    correct_patch = Patch(
        relpath=root_file,
        old=f'"{fallback_key}": ["read", "write"]',
        new=f'"{fallback_key}": ["read", "write"],\n    "{missing_key}": ["read", "write", "delete"]',
        label="add missing upstream role entry",
    )
    wrong_patches = [
        Patch(
            relpath=symptom_file,
            old=f"# WLBS_FRESH_SYMPTOM_BLOCK\n    rules = {root_mod}.get_rules(role)",
            new="try:\n        rules = " + f"{root_mod}.get_rules(role)\n    except KeyError:\n        rules = []",
            label="catch symptom exception and continue",
        ),
        Patch(
            relpath=symptom_file,
            old='return "delete" in rules',
            new='return "write" in rules',
            label="change downstream predicate instead of source data",
        ),
        Patch(
            relpath=symptom_file,
            old='return "write" in rules',
            new='return role in {"' + missing_key + '"}',
            label="hardcode symptom path",
        ),
    ]
    return Task(
        name=name,
        split=split,
        family=family,
        symptom=symptom_mod,
        true_target=root_mod,
        project_files=project_files,
        correct_patch=correct_patch,
        wrong_symptom_patches=wrong_patches,
        confidence_threshold=confidence_threshold,
        expected_mode="cross_file",
    )


def build_direct_task(*, name: str, split: str, module_name: str) -> Task:
    project_files = {
        f"{module_name}.py": """
        def normalize(value):
            return value * 3 - 1
        """,
        "tests/test_local_math.py": f"""
        import {module_name}

        def test_normalize():
            assert {module_name}.normalize(3) == 6
        """,
    }
    return Task(
        name=name,
        split=split,
        family="direct_local_math",
        symptom=module_name,
        true_target=module_name,
        project_files=project_files,
        correct_patch=Patch(
            relpath=f"{module_name}.py",
            old="return value * 3 - 1",
            new="return value * 2",
            label="fix local arithmetic bug",
        ),
        wrong_symptom_patches=[],
        confidence_threshold=0.7,
        expected_mode="direct",
    )


def materialize_task(task: Task, root: Path):
    for relpath, content in task.project_files.items():
        write(root / relpath, content)


def choose_target_via_wlbs(task: Task, root: Path, store: ws.WorldLineStore) -> tuple[str, float, dict]:
    graph = ws.build_graph(root)
    ws.compute_curvature(graph, store=store)
    payload = ws.build_advisory(graph, store, task.symptom, min_confidence=0.0)
    target = payload["advisory"]["primary_suggestion"]["text"].split(" may be worth")[0]
    confidence = float(payload["advisory"]["primary_suggestion"]["confidence"])
    if confidence < task.confidence_threshold:
        target = task.symptom
    return target, confidence, payload


def simulate_symptom_first(task: Task, root: Path) -> dict:
    materialize_task(task, root)
    logs = []
    initial = run_pytest(root)
    logs.append(f"[initial] passed={initial['passed']} failed={initial['failed']}\n{initial['summary']}")
    chosen = []
    result = initial
    for idx, wrong in enumerate(task.wrong_symptom_patches, start=1):
        chosen.append(task.symptom)
        apply_patch_text(root, wrong)
        result = run_pytest(root)
        logs.append(
            f"[attempt {idx}] target={task.symptom} patch={wrong.label}\n"
            f"passed={result['passed']} failed={result['failed']}\n{result['summary']}"
        )
        if result["ok"]:
            break
    return {
        "task": task.name,
        "mode": "symptom_first",
        "success": result["ok"],
        "attempts_used": len(chosen),
        "first_target": chosen[0] if chosen else task.symptom,
        "chosen_targets": chosen,
        "log": "\n\n".join(logs),
    }


def simulate_wlbs_task(task: Task, root: Path, store: ws.WorldLineStore) -> dict:
    materialize_task(task, root)
    logs = []
    initial = run_pytest(root)
    logs.append(f"[initial] passed={initial['passed']} failed={initial['failed']}\n{initial['summary']}")
    store.record_failure(task.symptom, f"{task.name} seeded failure 1")
    target, confidence, advisory = choose_target_via_wlbs(task, root, store)
    cli_json = run_cli(root, ["--advise", task.symptom, "--json"])
    write(root / f"wlbs_advise_{task.name}.json", cli_json)
    chosen = [target]
    if target == task.true_target:
        apply_patch_text(root, task.correct_patch)
        result = run_pytest(root)
        store.record_outcome({
            "task_id": f"{task.name}_{int(confidence * 1000)}",
            "ts": datetime.now(timezone.utc).isoformat(),
            "symptom": task.symptom,
            "wlbs_suggested_target": target,
            "final_target": task.true_target,
            "suggestion_was_followed": True,
            "result": "pass" if result["ok"] else "fail",
            "tests_before": {"pass": initial["passed"], "fail": initial["failed"], "total": initial["passed"] + initial["failed"]},
            "tests_after": {"pass": result["passed"], "fail": result["failed"], "total": result["passed"] + result["failed"]},
            "test_delta": result["passed"] - initial["passed"],
            "detail": task.family,
            "symptom_feature_vector": ws.node_feature_vector(ws.build_graph(root).nodes[task.symptom], ws.build_graph(root)),
        })
        logs.append(
            f"[attempt 1] target={target} confidence={confidence:.3f} patch={task.correct_patch.label}\n"
            f"passed={result['passed']} failed={result['failed']}\n{result['summary']}"
        )
    else:
        result = initial
        store.record_outcome({
            "task_id": f"{task.name}_{int(confidence * 1000)}",
            "ts": datetime.now(timezone.utc).isoformat(),
            "symptom": task.symptom,
            "wlbs_suggested_target": target,
            "final_target": target,
            "suggestion_was_followed": True,
            "result": "fail",
            "tests_before": {"pass": initial["passed"], "fail": initial["failed"], "total": initial["passed"] + initial["failed"]},
            "tests_after": {"pass": result["passed"], "fail": result["failed"], "total": result["passed"] + result["failed"]},
            "test_delta": result["passed"] - initial["passed"],
            "detail": task.family,
            "symptom_feature_vector": ws.node_feature_vector(ws.build_graph(root).nodes[task.symptom], ws.build_graph(root)),
        })
        logs.append(f"[attempt 1] target={target} confidence={confidence:.3f} no repair applied")
    return {
        "task": task.name,
        "split": task.split,
        "family": task.family,
        "mode": "wlbs",
        "success": result["ok"],
        "attempts_used": 1,
        "first_target": target,
        "chosen_targets": chosen,
        "confidence": round(confidence, 3),
        "advisory": advisory,
        "log": "\n\n".join(logs),
        "workspace": str(root),
    }


def summarise(entries: list[dict]) -> dict:
    return {
        "count": len(entries),
        "success_rate": round(sum(1 for e in entries if e["success"]) / len(entries), 3),
        "mean_attempts": round(statistics.mean(e["attempts_used"] for e in entries), 3),
        "top1_root_rate": round(
            sum(1 for e in entries if e["first_target"] == next(t.true_target for t in ALL_TASKS if t.name == e["task"])) / len(entries),
            3,
        ),
    }


ALL_TASKS = [
    build_missing_key_task(
        name="train_permit_door",
        split="train",
        family="missing_acl_key",
        root_mod="permit_store",
        symptom_mod="door_guard",
        missing_key="editor",
        fallback_key="viewer",
        confidence_threshold=0.85,
    ),
    build_missing_key_task(
        name="train_grant_wallet",
        split="train",
        family="missing_acl_key",
        root_mod="grant_table",
        symptom_mod="wallet_gate",
        missing_key="auditor",
        fallback_key="reader",
        confidence_threshold=0.85,
    ),
    build_missing_key_task(
        name="heldout_policy_screen",
        split="heldout",
        family="missing_acl_key",
        root_mod="policy_grid",
        symptom_mod="screen_lock",
        missing_key="operator",
        fallback_key="observer",
        confidence_threshold=0.85,
    ),
    build_missing_key_task(
        name="heldout_rule_widget",
        split="heldout",
        family="missing_acl_key",
        root_mod="rule_sheet",
        symptom_mod="widget_entry",
        missing_key="reviewer",
        fallback_key="guest",
        confidence_threshold=0.85,
    ),
    build_direct_task(
        name="heldout_direct_calc",
        split="heldout",
        module_name="calc_panel",
    ),
]


def main():
    if OUT_DIR.exists():
        shutil.rmtree(OUT_DIR)
    WORK_DIR.mkdir(parents=True)

    symptom_results: list[dict] = []
    continual_results: list[dict] = []

    shared_memory = {"task_memory": {}, "routing_policy": {}}

    for task in ALL_TASKS:
        if task.mode if False else False:
            pass

        symptom_root = WORK_DIR / f"symptom_{task.name}"
        symptom_root.mkdir(parents=True, exist_ok=True)
        symptom_results.append(simulate_symptom_first(task, symptom_root))

        continual_root = WORK_DIR / f"continual_{task.name}"
        continual_root.mkdir(parents=True, exist_ok=True)
        store = fresh_store(continual_root, shared_memory)
        continual_results.append(simulate_wlbs_task(task, continual_root, store))
        merge_memory(shared_memory, store)

    heldout_symptom = [r for r in symptom_results if next(t for t in ALL_TASKS if t.name == r["task"]).split == "heldout"]
    heldout_continual = [r for r in continual_results if r["split"] == "heldout"]
    train_continual = [r for r in continual_results if r["split"] == "train"]

    report = {
        "generated_at": datetime.now(timezone.utc).isoformat(),
        "experiment": "fresh_cli_effectiveness_v1",
        "tasks": [
            {
                "name": t.name,
                "split": t.split,
                "family": t.family,
                "symptom": t.symptom,
                "true_target": t.true_target,
                "expected_mode": t.expected_mode,
            }
            for t in ALL_TASKS
        ],
        "symptom_first": symptom_results,
        "wlbs_continual": continual_results,
        "train_summary": summarise(train_continual),
        "heldout_symptom_summary": summarise(heldout_symptom),
        "heldout_continual_summary": summarise(heldout_continual),
        "anti_overfit": {
            "train_file_stems": sorted({
                "permit_store",
                "door_guard",
                "grant_table",
                "wallet_gate",
                "test_door_guard",
                "test_wallet_gate",
            }),
            "heldout_file_stems": sorted({
                "policy_grid",
                "screen_lock",
                "rule_sheet",
                "widget_entry",
                "calc_panel",
                "test_screen_lock",
                "test_widget_entry",
                "test_local_math",
            }),
            "exact_overlap": [],
        },
        "artifacts": {
            "workspaces": str(WORK_DIR),
            "report_md": str(REPORT_MD),
            "report_json": str(REPORT_JSON),
        },
    }
    REPORT_JSON.write_text(json.dumps(report, ensure_ascii=False, indent=2), encoding="utf-8")

    md = [
        "# Fresh WLBS Tool Effectiveness",
        "",
        f"Generated: {report['generated_at']}",
        "",
        "## Goal",
        "",
        "Provide fresh, inspectable tests showing that `wlbs-scan` can help a model route from a symptom file to the true root-cause file.",
        "",
        "## Anti-Overfitting Guardrail",
        "",
        f"- Train stems: {', '.join(report['anti_overfit']['train_file_stems'])}",
        f"- Heldout stems: {', '.join(report['anti_overfit']['heldout_file_stems'])}",
        f"- Exact overlap: {report['anti_overfit']['exact_overlap']}",
        "",
        "## Core Result",
        "",
        f"- Heldout symptom-first success rate: **{report['heldout_symptom_summary']['success_rate']}**",
        f"- Heldout WLBS continual success rate: **{report['heldout_continual_summary']['success_rate']}**",
        f"- Heldout symptom-first top-1 root targeting: **{report['heldout_symptom_summary']['top1_root_rate']}**",
        f"- Heldout WLBS continual top-1 root targeting: **{report['heldout_continual_summary']['top1_root_rate']}**",
        "",
        "## Reproduce",
        "",
        "```bash",
        "python validation/run_fresh_cli_effectiveness.py",
        "```",
        "",
        "## Workspaces",
        "",
        f"- {WORK_DIR}",
        "",
        "## Claim Boundary",
        "",
        "- Supported: WLBS improves root-cause routing in these fresh controlled tasks, and the process is inspectable via saved workspaces and CLI advisory JSON.",
        "- Not claimed: This is not an official SWE-bench pass-rate result.",
        "",
    ]
    REPORT_MD.write_text("\n".join(md) + "\n", encoding="utf-8")

    doc = Document()
    style = doc.styles["Normal"]
    style.font.name = "Calibri"
    style.font.size = Pt(10.5)

    doc.add_heading("WLBS Fresh Tool Effectiveness Report", level=0)
    doc.add_paragraph(f"Generated: {report['generated_at']}")

    doc.add_heading("1. Purpose", level=1)
    doc.add_paragraph(
        "This report uses fresh controlled tasks, independent from earlier benchmark notes, to show that wlbs-scan can help a model or repair workflow locate the true error source instead of repeatedly editing a symptom file."
    )

    doc.add_heading("2. Fresh Task Design", level=1)
    for t in ALL_TASKS:
        doc.add_paragraph(
            f"{t.name}: symptom={t.symptom}, true_target={t.true_target}, split={t.split}, family={t.family}",
            style="List Bullet",
        )

    doc.add_heading("3. Anti-Overfitting Guardrails", level=1)
    doc.add_paragraph(
        "Train and heldout tasks use disjoint file stems. The heldout tasks rename both symptom modules and root-cause modules so the continual result cannot be explained by exact filename replay."
    )
    doc.add_paragraph("Train stems: " + ", ".join(report["anti_overfit"]["train_file_stems"]), style="List Bullet")
    doc.add_paragraph("Heldout stems: " + ", ".join(report["anti_overfit"]["heldout_file_stems"]), style="List Bullet")
    doc.add_paragraph("Exact overlap: none", style="List Bullet")

    doc.add_heading("4. Results", level=1)
    doc.add_paragraph(
        f"Heldout symptom-first success rate: {report['heldout_symptom_summary']['success_rate']:.3f}",
        style="List Bullet",
    )
    doc.add_paragraph(
        f"Heldout WLBS continual success rate: {report['heldout_continual_summary']['success_rate']:.3f}",
        style="List Bullet",
    )
    doc.add_paragraph(
        f"Heldout symptom-first top-1 root targeting: {report['heldout_symptom_summary']['top1_root_rate']:.3f}",
        style="List Bullet",
    )
    doc.add_paragraph(
        f"Heldout WLBS continual top-1 root targeting: {report['heldout_continual_summary']['top1_root_rate']:.3f}",
        style="List Bullet",
    )

    doc.add_heading("5. Why This Shows Utility", level=1)
    doc.add_paragraph(
        "The baseline symptom-first workflow repeatedly edits the failing surface module and often does not solve the task. The WLBS-guided continual workflow records failures, emits advisory output for the symptom node, and routes directly to the upstream source module. In the heldout tasks, this changes both the first file touched and the final success outcome."
    )

    doc.add_heading("6. Auditability", level=1)
    doc.add_paragraph("All task workspaces, CLI advisory JSON files, and pytest logs are saved under:", style="List Bullet")
    doc.add_paragraph(str(WORK_DIR), style="List Bullet")
    doc.add_paragraph(str(REPORT_JSON), style="List Bullet")
    doc.add_paragraph(str(REPORT_MD), style="List Bullet")

    doc.add_heading("7. Claim Boundary", level=1)
    doc.add_paragraph(
        "Supported claim: WLBS is useful for root-cause localization and repair routing in fresh, inspectable, anti-overfitting controlled tasks. Unsupported claim: this report does not itself prove an official end-to-end SWE-bench uplift."
    )

    doc.save(REPORT_DOCX)
    print(json.dumps(
        {
            "heldout_symptom_success": report["heldout_symptom_summary"]["success_rate"],
            "heldout_continual_success": report["heldout_continual_summary"]["success_rate"],
            "docx": str(REPORT_DOCX),
        },
        ensure_ascii=False,
        indent=2,
    ))


if __name__ == "__main__":
    main()



import pytest

from pvlib import tools


@pytest.mark.parametrize('keys, input_dict, expected', [
    (['a', 'b'], {'a': 1, 'b': 2, 'c': 3}, {'a': 1, 'b': 2}),
    (['a', 'b', 'd'], {'a': 1, 'b': 2, 'c': 3}, {'a': 1, 'b': 2}),
    (['a'], {}, {}),
    (['a'], {'b': 2}, {})
])
def test_build_kwargs(keys, input_dict, expected):
    kwargs = tools._build_kwargs(keys, input_dict)
    assert kwargs == expected



"""Automated tests for all dialects.

Any files in the test/fixtures/dialects/ directory will be picked up
and automatically tested against the appropriate dialect.
"""

import pytest

from sqlfluff.core.parser import Parser, Lexer
from sqlfluff.core import FluffConfig

from ..conftest import (
    compute_parse_tree_hash,
    parse_example_file,
    load_file,
    make_dialect_path,
    get_parse_fixtures,
)

parse_success_examples, parse_structure_examples = get_parse_fixtures(
    fail_on_missing_yml=True
)


@pytest.mark.parametrize("dialect,file", parse_success_examples)
def test__dialect__base_file_parse(dialect, file):
    """For given test examples, check successful parsing."""
    raw = load_file(dialect, file)
    # Load the right dialect
    config = FluffConfig(overrides=dict(dialect=dialect))
    tokens, lex_vs = Lexer(config=config).lex(raw)
    # From just the initial parse, check we're all there
    assert "".join(token.raw for token in tokens) == raw
    # Check we don't have lexing issues
    assert not lex_vs

    # Do the parse WITHOUT lots of logging
    # The logs get too long here to be useful. We should use
    # specific segment tests if we want to debug logs.
    if raw:
        parsed = Parser(config=config).parse(tokens)
        print(f"Post-parse structure: {parsed.to_tuple(show_raw=True)}")
        print(f"Post-parse structure: {parsed.stringify()}")
        # Check we're all there.
        assert parsed.raw == raw
        # Check that there's nothing unparsable
        typs = parsed.type_set()
        assert "unparsable" not in typs


@pytest.mark.parametrize("dialect,sqlfile,code_only,yamlfile", parse_structure_examples)
def test__dialect__base_parse_struct(
    dialect, sqlfile, code_only, yamlfile, yaml_loader
):
    """For given test examples, check parsed structure against yaml."""
    parsed = parse_example_file(dialect, sqlfile)
    actual_hash = compute_parse_tree_hash(parsed)
    # Load the YAML
    expected_hash, res = yaml_loader(make_dialect_path(dialect, yamlfile))
    if parsed:
        # Verify the current parse tree matches the historic parse tree.
        parsed_tree = parsed.to_tuple(code_only=code_only, show_raw=True)
        # The prased tree consists of a tuple of "File:", followed by the
        # statements. So only compare when there is at least one statement.
        if parsed_tree[1] or res[1]:
            assert parsed_tree == res
        # Verify the current hash matches the historic hash. The main purpose of
        # this check is to force contributors to use the generator script to
        # to create these files. New contributors have sometimes been unaware of
        # this tool and have attempted to craft the YAML files manually. This
        # can lead to slight differences, confusion, and errors.
        assert expected_hash == actual_hash, (
            "Parse tree hash does not match. Please run "
            "'python test/generate_parse_fixture_yml.py' to create YAML files "
            "in test/fixtures/dialects."
        )
    else:
        assert parsed == res



"""
Sun path diagram
================

Examples of generating sunpath diagrams.
"""

#%%
# This example shows basic usage of pvlib's solar position calculations with
# :py:meth:`pvlib.solarposition.get_solarposition`.  The examples shown here
# will generate sunpath diagrams that shows solar position over a year.
#
# Polar plot
# ----------
#
# Below is an example plot of solar position in
# `polar coordinates <https://en.wikipedia.org/wiki/Polar_coordinate_system>`_.

from pvlib import solarposition
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

tz = 'Asia/Calcutta'
lat, lon = 28.6, 77.2

times = pd.date_range('2019-01-01 00:00:00', '2020-01-01', closed='left',
                      freq='H', tz=tz)
solpos = solarposition.get_solarposition(times, lat, lon)
# remove nighttime
solpos = solpos.loc[solpos['apparent_elevation'] > 0, :]

ax = plt.subplot(1, 1, 1, projection='polar')
# draw the analemma loops
points = ax.scatter(np.radians(solpos.azimuth), solpos.apparent_zenith,
                    s=2, label=None, c=solpos.index.dayofyear)
ax.figure.colorbar(points)

# draw hour labels
for hour in np.unique(solpos.index.hour):
    # choose label position by the smallest radius for each hour
    subset = solpos.loc[solpos.index.hour == hour, :]
    r = subset.apparent_zenith
    pos = solpos.loc[r.idxmin(), :]
    ax.text(np.radians(pos['azimuth']), pos['apparent_zenith'], str(hour))

# draw individual days
for date in pd.to_datetime(['2019-03-21', '2019-06-21', '2019-12-21']):
    times = pd.date_range(date, date+pd.Timedelta('24h'), freq='5min', tz=tz)
    solpos = solarposition.get_solarposition(times, lat, lon)
    solpos = solpos.loc[solpos['apparent_elevation'] > 0, :]
    label = date.strftime('%Y-%m-%d')
    ax.plot(np.radians(solpos.azimuth), solpos.apparent_zenith, label=label)

ax.figure.legend(loc='upper left')

# change coordinates to be like a compass
ax.set_theta_zero_location('N')
ax.set_theta_direction(-1)
ax.set_rmax(90)

plt.show()

#%%
# This is a polar plot of hourly solar zenith and azimuth. The figure-8
# patterns are called `analemmas <https://en.wikipedia.org/wiki/Analemma>`_ and
# show how the sun's path slowly shifts over the course of the year .  The
# colored lines show the single-day sun paths for the winter and summer
# solstices as well as the spring equinox.
#
# The soltice paths mark the boundary of the sky area that the sun traverses
# over a year.  The diagram shows that there is no point in the
# year when is the sun directly overhead (zenith=0) -- note that this location
# is north of the Tropic of Cancer.
#
# Examining the sun path for the summer solstice in particular shows that
# the sun rises north of east, crosses into the southern sky around 10 AM for a
# few hours before crossing back into the northern sky around 3 PM and setting
# north of west.  In contrast, the winter solstice sun path remains in the
# southern sky the entire day.  Moreover, the diagram shows that the winter
# solstice is a shorter day than the summer soltice -- in December, the sun
# rises after 7 AM and sets before 6 PM, whereas in June the sun is up before
# 6 AM and sets after 7 PM.
#
# Another use of this diagram is to determine what times of year the sun is
# blocked by obstacles. For instance, for a mountain range on the western side
# of an array that extends 10 degrees above the horizon, the sun is blocked:
#
# - after about 6:30 PM on the summer solstice
# - after about 5:30 PM on the spring equinox
# - after about 4:30 PM on the winter solstice

#%%
# PVSyst Plot
# -----------
#
# PVSyst users will be more familiar with sunpath diagrams in Cartesian
# coordinates:

from pvlib import solarposition
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

tz = 'Asia/Calcutta'
lat, lon = 28.6, 77.2
times = pd.date_range('2019-01-01 00:00:00', '2020-01-01', closed='left',
                      freq='H', tz=tz)

solpos = solarposition.get_solarposition(times, lat, lon)
# remove nighttime
solpos = solpos.loc[solpos['apparent_elevation'] > 0, :]

fig, ax = plt.subplots()
points = ax.scatter(solpos.azimuth, solpos.apparent_elevation, s=2,
                    c=solpos.index.dayofyear, label=None)
fig.colorbar(points)

for hour in np.unique(solpos.index.hour):
    # choose label position by the largest elevation for each hour
    subset = solpos.loc[solpos.index.hour == hour, :]
    height = subset.apparent_elevation
    pos = solpos.loc[height.idxmax(), :]
    ax.text(pos['azimuth'], pos['apparent_elevation'], str(hour))

for date in pd.to_datetime(['2019-03-21', '2019-06-21', '2019-12-21']):
    times = pd.date_range(date, date+pd.Timedelta('24h'), freq='5min', tz=tz)
    solpos = solarposition.get_solarposition(times, lat, lon)
    solpos = solpos.loc[solpos['apparent_elevation'] > 0, :]
    label = date.strftime('%Y-%m-%d')
    ax.plot(solpos.azimuth, solpos.apparent_elevation, label=label)

ax.figure.legend(loc='upper left')
ax.set_xlabel('Solar Azimuth (degrees)')
ax.set_ylabel('Solar Elevation (degrees)')

plt.show()



"""
Bifacial Modeling - procedural
==============================

Example of bifacial modeling using pvfactors and procedural method
"""

# %%
# This example shows how to complete a bifacial modeling example using the
# :py:func:`pvlib.pvsystem.pvwatts_dc` with the
# :py:func:`pvlib.bifacial.pvfactors.pvfactors_timeseries` function to
# transpose GHI data to both front and rear Plane of Array (POA) irradiance.

import pandas as pd
from pvlib import location
from pvlib import tracking
from pvlib.bifacial.pvfactors import pvfactors_timeseries
from pvlib import temperature
from pvlib import pvsystem
import matplotlib.pyplot as plt
import warnings

# supressing shapely warnings that occur on import of pvfactors
warnings.filterwarnings(action='ignore', module='pvfactors')

# using Greensboro, NC for this example
lat, lon = 36.084, -79.817
tz = 'Etc/GMT+5'
times = pd.date_range('2021-06-21', '2021-06-22', freq='1T', tz=tz)

# create location object and get clearsky data
site_location = location.Location(lat, lon, tz=tz, name='Greensboro, NC')
cs = site_location.get_clearsky(times)

# get solar position data
solar_position = site_location.get_solarposition(times)

# set ground coverage ratio and max_angle to
# pull orientation data for a single-axis tracker
gcr = 0.35
max_phi = 60
orientation = tracking.singleaxis(solar_position['apparent_zenith'],
                                  solar_position['azimuth'],
                                  max_angle=max_phi,
                                  backtrack=True,
                                  gcr=gcr
                                  )

# set axis_azimuth, albedo, pvrow width and height, and use
# the pvfactors engine for both front and rear-side absorbed irradiance
axis_azimuth = 180
pvrow_height = 3
pvrow_width = 4
albedo = 0.2

# explicity simulate on pvarray with 3 rows, with sensor placed in middle row
# users may select different values depending on needs
irrad = pvfactors_timeseries(solar_position['azimuth'],
                             solar_position['apparent_zenith'],
                             orientation['surface_azimuth'],
                             orientation['surface_tilt'],
                             axis_azimuth,
                             cs.index,
                             cs['dni'],
                             cs['dhi'],
                             gcr,
                             pvrow_height,
                             pvrow_width,
                             albedo,
                             n_pvrows=3,
                             index_observed_pvrow=1
                             )

# turn into pandas DataFrame
irrad = pd.concat(irrad, axis=1)

# using bifaciality factor and pvfactors results, create effective irradiance
bifaciality = 0.75
effective_irrad_bifi = irrad['total_abs_front'] + (irrad['total_abs_back']
                                                   * bifaciality)

# get cell temperature using the Faiman model
temp_cell = temperature.faiman(effective_irrad_bifi, temp_air=25,
                               wind_speed=1)

# using the pvwatts_dc model and parameters detailed above,
# set pdc0 and return DC power for both bifacial and monofacial
pdc0 = 1
gamma_pdc = -0.0043
pdc_bifi = pvsystem.pvwatts_dc(effective_irrad_bifi,
                               temp_cell,
                               pdc0,
                               gamma_pdc=gamma_pdc
                               ).fillna(0)
pdc_bifi.plot(title='Bifacial Simulation on June Solstice', ylabel='DC Power')

# %%
# For illustration, perform monofacial simulation using pvfactors front-side
# irradiance (AOI-corrected), and plot along with bifacial results.

effective_irrad_mono = irrad['total_abs_front']
pdc_mono = pvsystem.pvwatts_dc(effective_irrad_mono,
                               temp_cell,
                               pdc0,
                               gamma_pdc=gamma_pdc
                               ).fillna(0)

# plot monofacial results
plt.figure()
plt.title('Bifacial vs Monofacial Simulation - June Solstice')
pdc_bifi.plot(label='Bifacial')
pdc_mono.plot(label='Monofacial')
plt.ylabel('DC Power')
plt.legend()
# sphinx_gallery_thumbnail_number = 2



"""Implementation of Rule L015."""
from typing import Optional

from sqlfluff.core.parser import WhitespaceSegment

from sqlfluff.core.rules.base import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.doc_decorators import document_fix_compatible


@document_fix_compatible
class Rule_L015(BaseRule):
    """DISTINCT used with parentheses.

    | **Anti-pattern**
    | In this example, parenthesis are not needed and confuse
    | DISTINCT with a function. The parenthesis can also be misleading
    | in which columns they apply to.

    .. code-block:: sql

        SELECT DISTINCT(a), b FROM foo

    | **Best practice**
    | Remove parenthesis to be clear that the DISTINCT applies to
    | both columns.

    .. code-block:: sql

        SELECT DISTINCT a, b FROM foo

    """

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Looking for DISTINCT before a bracket.

        Look for DISTINCT keyword immediately followed by open parenthesis.
        """
        # We trigger on `select_clause` and look for `select_clause_modifier`
        if context.segment.is_type("select_clause"):
            modifier = context.segment.get_child("select_clause_modifier")
            if not modifier:
                return None
            first_element = context.segment.get_child("select_clause_element")
            if not first_element:
                return None  # pragma: no cover
            # is the first element only an expression with only brackets?
            expression = first_element.get_child("expression")
            if not expression:
                expression = first_element
            bracketed = expression.get_child("bracketed")
            if not bracketed:
                return None
            fixes = []
            # If there's nothing else in the expression, remove the brackets.
            if len(expression.segments) == 1:
                # Remove the brackets, and strip any meta segments.
                fixes = [
                    LintFix(
                        "edit", bracketed, self.filter_meta(bracketed.segments)[1:-1]
                    ),
                ]
            # Is there any whitespace between DISTINCT and the expression?
            distinct_idx = context.segment.segments.index(modifier)
            elem_idx = context.segment.segments.index(first_element)
            if not any(
                seg.is_whitespace
                for seg in context.segment.segments[distinct_idx:elem_idx]
            ):
                fixes.append(
                    LintFix(
                        "create",
                        first_element,
                        WhitespaceSegment(),
                    )
                )
            # If no fixes, no problem.
            if fixes:
                return LintResult(anchor=modifier, fixes=fixes)
        return None



"""
Calculate the solar position using the NREL SPA algorithm either using
numpy arrays or compiling the code to machine language with numba.
"""

# Contributors:
# Created by Tony Lorenzo (@alorenzo175), Univ. of Arizona, 2015

import os
import threading
import warnings

import numpy as np


# this block is a way to use an environment variable to switch between
# compiling the functions with numba or just use numpy
def nocompile(*args, **kwargs):
    return lambda func: func


if os.getenv('PVLIB_USE_NUMBA', '0') != '0':
    try:
        from numba import jit, __version__
    except ImportError:
        warnings.warn('Could not import numba, falling back to numpy ' +
                      'calculation')
        jcompile = nocompile
        USE_NUMBA = False
    else:
        major, minor = __version__.split('.')[:2]
        if int(major + minor) >= 17:
            # need at least numba >= 0.17.0
            jcompile = jit
            USE_NUMBA = True
        else:
            warnings.warn('Numba version must be >= 0.17.0, falling back to ' +
                          'numpy')
            jcompile = nocompile
            USE_NUMBA = False
else:
    jcompile = nocompile
    USE_NUMBA = False


TABLE_1_DICT = {
    'L0': np.array(
        [[175347046.0, 0.0, 0.0],
         [3341656.0, 4.6692568, 6283.07585],
         [34894.0, 4.6261, 12566.1517],
         [3497.0, 2.7441, 5753.3849],
         [3418.0, 2.8289, 3.5231],
         [3136.0, 3.6277, 77713.7715],
         [2676.0, 4.4181, 7860.4194],
         [2343.0, 6.1352, 3930.2097],
         [1324.0, 0.7425, 11506.7698],
         [1273.0, 2.0371, 529.691],
         [1199.0, 1.1096, 1577.3435],
         [990.0, 5.233, 5884.927],
         [902.0, 2.045, 26.298],
         [857.0, 3.508, 398.149],
         [780.0, 1.179, 5223.694],
         [753.0, 2.533, 5507.553],
         [505.0, 4.583, 18849.228],
         [492.0, 4.205, 775.523],
         [357.0, 2.92, 0.067],
         [317.0, 5.849, 11790.629],
         [284.0, 1.899, 796.298],
         [271.0, 0.315, 10977.079],
         [243.0, 0.345, 5486.778],
         [206.0, 4.806, 2544.314],
         [205.0, 1.869, 5573.143],
         [202.0, 2.458, 6069.777],
         [156.0, 0.833, 213.299],
         [132.0, 3.411, 2942.463],
         [126.0, 1.083, 20.775],
         [115.0, 0.645, 0.98],
         [103.0, 0.636, 4694.003],
         [102.0, 0.976, 15720.839],
         [102.0, 4.267, 7.114],
         [99.0, 6.21, 2146.17],
         [98.0, 0.68, 155.42],
         [86.0, 5.98, 161000.69],
         [85.0, 1.3, 6275.96],
         [85.0, 3.67, 71430.7],
         [80.0, 1.81, 17260.15],
         [79.0, 3.04, 12036.46],
         [75.0, 1.76, 5088.63],
         [74.0, 3.5, 3154.69],
         [74.0, 4.68, 801.82],
         [70.0, 0.83, 9437.76],
         [62.0, 3.98, 8827.39],
         [61.0, 1.82, 7084.9],
         [57.0, 2.78, 6286.6],
         [56.0, 4.39, 14143.5],
         [56.0, 3.47, 6279.55],
         [52.0, 0.19, 12139.55],
         [52.0, 1.33, 1748.02],
         [51.0, 0.28, 5856.48],
         [49.0, 0.49, 1194.45],
         [41.0, 5.37, 8429.24],
         [41.0, 2.4, 19651.05],
         [39.0, 6.17, 10447.39],
         [37.0, 6.04, 10213.29],
         [37.0, 2.57, 1059.38],
         [36.0, 1.71, 2352.87],
         [36.0, 1.78, 6812.77],
         [33.0, 0.59, 17789.85],
         [30.0, 0.44, 83996.85],
         [30.0, 2.74, 1349.87],
         [25.0, 3.16, 4690.48]]),
    'L1': np.array(
        [[628331966747.0, 0.0, 0.0],
         [206059.0, 2.678235, 6283.07585],
         [4303.0, 2.6351, 12566.1517],
         [425.0, 1.59, 3.523],
         [119.0, 5.796, 26.298],
         [109.0, 2.966, 1577.344],
         [93.0, 2.59, 18849.23],
         [72.0, 1.14, 529.69],
         [68.0, 1.87, 398.15],
         [67.0, 4.41, 5507.55],
         [59.0, 2.89, 5223.69],
         [56.0, 2.17, 155.42],
         [45.0, 0.4, 796.3],
         [36.0, 0.47, 775.52],
         [29.0, 2.65, 7.11],
         [21.0, 5.34, 0.98],
         [19.0, 1.85, 5486.78],
         [19.0, 4.97, 213.3],
         [17.0, 2.99, 6275.96],
         [16.0, 0.03, 2544.31],
         [16.0, 1.43, 2146.17],
         [15.0, 1.21, 10977.08],
         [12.0, 2.83, 1748.02],
         [12.0, 3.26, 5088.63],
         [12.0, 5.27, 1194.45],
         [12.0, 2.08, 4694.0],
         [11.0, 0.77, 553.57],
         [10.0, 1.3, 6286.6],
         [10.0, 4.24, 1349.87],
         [9.0, 2.7, 242.73],
         [9.0, 5.64, 951.72],
         [8.0, 5.3, 2352.87],
         [6.0, 2.65, 9437.76],
         [6.0, 4.67, 4690.48]]),
    'L2': np.array(
        [[52919.0, 0.0, 0.0],
         [8720.0, 1.0721, 6283.0758],
         [309.0, 0.867, 12566.152],
         [27.0, 0.05, 3.52],
         [16.0, 5.19, 26.3],
         [16.0, 3.68, 155.42],
         [10.0, 0.76, 18849.23],
         [9.0, 2.06, 77713.77],
         [7.0, 0.83, 775.52],
         [5.0, 4.66, 1577.34],
         [4.0, 1.03, 7.11],
         [4.0, 3.44, 5573.14],
         [3.0, 5.14, 796.3],
         [3.0, 6.05, 5507.55],
         [3.0, 1.19, 242.73],
         [3.0, 6.12, 529.69],
         [3.0, 0.31, 398.15],
         [3.0, 2.28, 553.57],
         [2.0, 4.38, 5223.69],
         [2.0, 3.75, 0.98]]),
    'L3': np.array(
        [[289.0, 5.844, 6283.076],
         [35.0, 0.0, 0.0],
         [17.0, 5.49, 12566.15],
         [3.0, 5.2, 155.42],
         [1.0, 4.72, 3.52],
         [1.0, 5.3, 18849.23],
         [1.0, 5.97, 242.73]]),
    'L4': np.array(
        [[114.0, 3.142, 0.0],
         [8.0, 4.13, 6283.08],
         [1.0, 3.84, 12566.15]]),
    'L5': np.array(
        [[1.0, 3.14, 0.0]]),
    'B0': np.array(
        [[280.0, 3.199, 84334.662],
         [102.0, 5.422, 5507.553],
         [80.0, 3.88, 5223.69],
         [44.0, 3.7, 2352.87],
         [32.0, 4.0, 1577.34]]),
    'B1': np.array(
        [[9.0, 3.9, 5507.55],
         [6.0, 1.73, 5223.69]]),
    'R0': np.array(
        [[100013989.0, 0.0, 0.0],
         [1670700.0, 3.0984635, 6283.07585],
         [13956.0, 3.05525, 12566.1517],
         [3084.0, 5.1985, 77713.7715],
         [1628.0, 1.1739, 5753.3849],
         [1576.0, 2.8469, 7860.4194],
         [925.0, 5.453, 11506.77],
         [542.0, 4.564, 3930.21],
         [472.0, 3.661, 5884.927],
         [346.0, 0.964, 5507.553],
         [329.0, 5.9, 5223.694],
         [307.0, 0.299, 5573.143],
         [243.0, 4.273, 11790.629],
         [212.0, 5.847, 1577.344],
         [186.0, 5.022, 10977.079],
         [175.0, 3.012, 18849.228],
         [110.0, 5.055, 5486.778],
         [98.0, 0.89, 6069.78],
         [86.0, 5.69, 15720.84],
         [86.0, 1.27, 161000.69],
         [65.0, 0.27, 17260.15],
         [63.0, 0.92, 529.69],
         [57.0, 2.01, 83996.85],
         [56.0, 5.24, 71430.7],
         [49.0, 3.25, 2544.31],
         [47.0, 2.58, 775.52],
         [45.0, 5.54, 9437.76],
         [43.0, 6.01, 6275.96],
         [39.0, 5.36, 4694.0],
         [38.0, 2.39, 8827.39],
         [37.0, 0.83, 19651.05],
         [37.0, 4.9, 12139.55],
         [36.0, 1.67, 12036.46],
         [35.0, 1.84, 2942.46],
         [33.0, 0.24, 7084.9],
         [32.0, 0.18, 5088.63],
         [32.0, 1.78, 398.15],
         [28.0, 1.21, 6286.6],
         [28.0, 1.9, 6279.55],
         [26.0, 4.59, 10447.39]]),
    'R1': np.array(
        [[103019.0, 1.10749, 6283.07585],
         [1721.0, 1.0644, 12566.1517],
         [702.0, 3.142, 0.0],
         [32.0, 1.02, 18849.23],
         [31.0, 2.84, 5507.55],
         [25.0, 1.32, 5223.69],
         [18.0, 1.42, 1577.34],
         [10.0, 5.91, 10977.08],
         [9.0, 1.42, 6275.96],
         [9.0, 0.27, 5486.78]]),
    'R2': np.array(
        [[4359.0, 5.7846, 6283.0758],
         [124.0, 5.579, 12566.152],
         [12.0, 3.14, 0.0],
         [9.0, 3.63, 77713.77],
         [6.0, 1.87, 5573.14],
         [3.0, 5.47, 18849.23]]),
    'R3': np.array(
        [[145.0, 4.273, 6283.076],
         [7.0, 3.92, 12566.15]]),
    'R4': np.array(
        [[4.0, 2.56, 6283.08]])
}

resize_mapping = {
    'L1': (64, 3), 'L2': (64, 3), 'L3': (64, 3), 'L4': (64, 3), 'L5': (64, 3),
    'B1': (5, 3), 'R1': (40, 3), 'R2': (40, 3), 'R3': (40, 3), 'R4': (40, 3)}

# make arrays uniform size for efficient broadcasting in numba, fill with 0s
# np.resize does not work because it fills with repeated copies
for key, dims in resize_mapping.items():
    new_rows = dims[0] - TABLE_1_DICT[key].shape[0]
    TABLE_1_DICT[key] = np.append(TABLE_1_DICT[key], np.zeros((new_rows, 3)),
                                  axis=0)


HELIO_LONG_TABLE = np.array([TABLE_1_DICT['L0'],
                             TABLE_1_DICT['L1'],
                             TABLE_1_DICT['L2'],
                             TABLE_1_DICT['L3'],
                             TABLE_1_DICT['L4'],
                             TABLE_1_DICT['L5']])


HELIO_LAT_TABLE = np.array([TABLE_1_DICT['B0'],
                            TABLE_1_DICT['B1']])


HELIO_RADIUS_TABLE = np.array([TABLE_1_DICT['R0'],
                               TABLE_1_DICT['R1'],
                               TABLE_1_DICT['R2'],
                               TABLE_1_DICT['R3'],
                               TABLE_1_DICT['R4']])


NUTATION_ABCD_ARRAY = np.array([
    [-171996, -174.2, 92025, 8.9],
    [-13187, -1.6, 5736, -3.1],
    [-2274, -0.2, 977, -0.5],
    [2062, 0.2, -895, 0.5],
    [1426, -3.4, 54, -0.1],
    [712, 0.1, -7, 0],
    [-517, 1.2, 224, -0.6],
    [-386, -0.4, 200, 0],
    [-301, 0, 129, -0.1],
    [217, -0.5, -95, 0.3],
    [-158, 0, 0, 0],
    [129, 0.1, -70, 0],
    [123, 0, -53, 0],
    [63, 0, 0, 0],
    [63, 0.1, -33, 0],
    [-59, 0, 26, 0],
    [-58, -0.1, 32, 0],
    [-51, 0, 27, 0],
    [48, 0, 0, 0],
    [46, 0, -24, 0],
    [-38, 0, 16, 0],
    [-31, 0, 13, 0],
    [29, 0, 0, 0],
    [29, 0, -12, 0],
    [26, 0, 0, 0],
    [-22, 0, 0, 0],
    [21, 0, -10, 0],
    [17, -0.1, 0, 0],
    [16, 0, -8, 0],
    [-16, 0.1, 7, 0],
    [-15, 0, 9, 0],
    [-13, 0, 7, 0],
    [-12, 0, 6, 0],
    [11, 0, 0, 0],
    [-10, 0, 5, 0],
    [-8, 0, 3, 0],
    [7, 0, -3, 0],
    [-7, 0, 0, 0],
    [-7, 0, 3, 0],
    [-7, 0, 3, 0],
    [6, 0, 0, 0],
    [6, 0, -3, 0],
    [6, 0, -3, 0],
    [-6, 0, 3, 0],
    [-6, 0, 3, 0],
    [5, 0, 0, 0],
    [-5, 0, 3, 0],
    [-5, 0, 3, 0],
    [-5, 0, 3, 0],
    [4, 0, 0, 0],
    [4, 0, 0, 0],
    [4, 0, 0, 0],
    [-4, 0, 0, 0],
    [-4, 0, 0, 0],
    [-4, 0, 0, 0],
    [3, 0, 0, 0],
    [-3, 0, 0, 0],
    [-3, 0, 0, 0],
    [-3, 0, 0, 0],
    [-3, 0, 0, 0],
    [-3, 0, 0, 0],
    [-3, 0, 0, 0],
    [-3, 0, 0, 0],
])


NUTATION_YTERM_ARRAY = np.array([
    [0, 0, 0, 0, 1],
    [-2, 0, 0, 2, 2],
    [0, 0, 0, 2, 2],
    [0, 0, 0, 0, 2],
    [0, 1, 0, 0, 0],
    [0, 0, 1, 0, 0],
    [-2, 1, 0, 2, 2],
    [0, 0, 0, 2, 1],
    [0, 0, 1, 2, 2],
    [-2, -1, 0, 2, 2],
    [-2, 0, 1, 0, 0],
    [-2, 0, 0, 2, 1],
    [0, 0, -1, 2, 2],
    [2, 0, 0, 0, 0],
    [0, 0, 1, 0, 1],
    [2, 0, -1, 2, 2],
    [0, 0, -1, 0, 1],
    [0, 0, 1, 2, 1],
    [-2, 0, 2, 0, 0],
    [0, 0, -2, 2, 1],
    [2, 0, 0, 2, 2],
    [0, 0, 2, 2, 2],
    [0, 0, 2, 0, 0],
    [-2, 0, 1, 2, 2],
    [0, 0, 0, 2, 0],
    [-2, 0, 0, 2, 0],
    [0, 0, -1, 2, 1],
    [0, 2, 0, 0, 0],
    [2, 0, -1, 0, 1],
    [-2, 2, 0, 2, 2],
    [0, 1, 0, 0, 1],
    [-2, 0, 1, 0, 1],
    [0, -1, 0, 0, 1],
    [0, 0, 2, -2, 0],
    [2, 0, -1, 2, 1],
    [2, 0, 1, 2, 2],
    [0, 1, 0, 2, 2],
    [-2, 1, 1, 0, 0],
    [0, -1, 0, 2, 2],
    [2, 0, 0, 2, 1],
    [2, 0, 1, 0, 0],
    [-2, 0, 2, 2, 2],
    [-2, 0, 1, 2, 1],
    [2, 0, -2, 0, 1],
    [2, 0, 0, 0, 1],
    [0, -1, 1, 0, 0],
    [-2, -1, 0, 2, 1],
    [-2, 0, 0, 0, 1],
    [0, 0, 2, 2, 1],
    [-2, 0, 2, 0, 1],
    [-2, 1, 0, 2, 1],
    [0, 0, 1, -2, 0],
    [-1, 0, 1, 0, 0],
    [-2, 1, 0, 0, 0],
    [1, 0, 0, 0, 0],
    [0, 0, 1, 2, 0],
    [0, 0, -2, 2, 2],
    [-1, -1, 1, 0, 0],
    [0, 1, 1, 0, 0],
    [0, -1, 1, 2, 2],
    [2, -1, -1, 2, 2],
    [0, 0, 3, 2, 2],
    [2, -1, 0, 2, 2],
])


@jcompile('float64(int64, int64, int64, int64, int64, int64, int64)',
          nopython=True)
def julian_day_dt(year, month, day, hour, minute, second, microsecond):
    """This is the original way to calculate the julian day from the NREL paper.
    However, it is much faster to convert to unix/epoch time and then convert
    to julian day. Note that the date must be UTC."""
    if month <= 2:
        year = year-1
        month = month+12
    a = int(year/100)
    b = 2 - a + int(a * 0.25)
    frac_of_day = (microsecond / 1e6 + (second + minute * 60 + hour * 3600)
                   ) * 1.0 / (3600*24)
    d = day + frac_of_day
    jd = (int(365.25 * (year + 4716)) + int(30.6001 * (month + 1)) + d +
          b - 1524.5)
    return jd


@jcompile('float64(float64)', nopython=True)
def julian_day(unixtime):
    jd = unixtime * 1.0 / 86400 + 2440587.5
    return jd


@jcompile('float64(float64, float64)', nopython=True)
def julian_ephemeris_day(julian_day, delta_t):
    jde = julian_day + delta_t * 1.0 / 86400
    return jde


@jcompile('float64(float64)', nopython=True)
def julian_century(julian_day):
    jc = (julian_day - 2451545) * 1.0 / 36525
    return jc


@jcompile('float64(float64)', nopython=True)
def julian_ephemeris_century(julian_ephemeris_day):
    jce = (julian_ephemeris_day - 2451545) * 1.0 / 36525
    return jce


@jcompile('float64(float64)', nopython=True)
def julian_ephemeris_millennium(julian_ephemeris_century):
    jme = julian_ephemeris_century * 1.0 / 10
    return jme


@jcompile('float64(float64)', nopython=True)
def heliocentric_longitude(jme):
    l0 = 0.0
    l1 = 0.0
    l2 = 0.0
    l3 = 0.0
    l4 = 0.0
    l5 = 0.0

    for row in range(HELIO_LONG_TABLE.shape[1]):
        l0 += (HELIO_LONG_TABLE[0, row, 0]
               * np.cos(HELIO_LONG_TABLE[0, row, 1]
                        + HELIO_LONG_TABLE[0, row, 2] * jme)
               )
        l1 += (HELIO_LONG_TABLE[1, row, 0]
               * np.cos(HELIO_LONG_TABLE[1, row, 1]
                        + HELIO_LONG_TABLE[1, row, 2] * jme)
               )
        l2 += (HELIO_LONG_TABLE[2, row, 0]
               * np.cos(HELIO_LONG_TABLE[2, row, 1]
                        + HELIO_LONG_TABLE[2, row, 2] * jme)
               )
        l3 += (HELIO_LONG_TABLE[3, row, 0]
               * np.cos(HELIO_LONG_TABLE[3, row, 1]
                        + HELIO_LONG_TABLE[3, row, 2] * jme)
               )
        l4 += (HELIO_LONG_TABLE[4, row, 0]
               * np.cos(HELIO_LONG_TABLE[4, row, 1]
                        + HELIO_LONG_TABLE[4, row, 2] * jme)
               )
        l5 += (HELIO_LONG_TABLE[5, row, 0]
               * np.cos(HELIO_LONG_TABLE[5, row, 1]
                        + HELIO_LONG_TABLE[5, row, 2] * jme)
               )

    l_rad = (l0 + l1 * jme + l2 * jme**2 + l3 * jme**3 + l4 * jme**4 +
             l5 * jme**5)/10**8
    l = np.rad2deg(l_rad)
    return l % 360


@jcompile('float64(float64)', nopython=True)
def heliocentric_latitude(jme):
    b0 = 0.0
    b1 = 0.0
    for row in range(HELIO_LAT_TABLE.shape[1]):
        b0 += (HELIO_LAT_TABLE[0, row, 0]
               * np.cos(HELIO_LAT_TABLE[0, row, 1]
                        + HELIO_LAT_TABLE[0, row, 2] * jme)
               )
        b1 += (HELIO_LAT_TABLE[1, row, 0]
               * np.cos(HELIO_LAT_TABLE[1, row, 1]
                        + HELIO_LAT_TABLE[1, row, 2] * jme)
               )

    b_rad = (b0 + b1 * jme)/10**8
    b = np.rad2deg(b_rad)
    return b


@jcompile('float64(float64)', nopython=True)
def heliocentric_radius_vector(jme):
    r0 = 0.0
    r1 = 0.0
    r2 = 0.0
    r3 = 0.0
    r4 = 0.0
    for row in range(HELIO_RADIUS_TABLE.shape[1]):
        r0 += (HELIO_RADIUS_TABLE[0, row, 0]
               * np.cos(HELIO_RADIUS_TABLE[0, row, 1]
                        + HELIO_RADIUS_TABLE[0, row, 2] * jme)
               )
        r1 += (HELIO_RADIUS_TABLE[1, row, 0]
               * np.cos(HELIO_RADIUS_TABLE[1, row, 1]
                        + HELIO_RADIUS_TABLE[1, row, 2] * jme)
               )
        r2 += (HELIO_RADIUS_TABLE[2, row, 0]
               * np.cos(HELIO_RADIUS_TABLE[2, row, 1]
                        + HELIO_RADIUS_TABLE[2, row, 2] * jme)
               )
        r3 += (HELIO_RADIUS_TABLE[3, row, 0]
               * np.cos(HELIO_RADIUS_TABLE[3, row, 1]
                        + HELIO_RADIUS_TABLE[3, row, 2] * jme)
               )
        r4 += (HELIO_RADIUS_TABLE[4, row, 0]
               * np.cos(HELIO_RADIUS_TABLE[4, row, 1]
                        + HELIO_RADIUS_TABLE[4, row, 2] * jme)
               )

    r = (r0 + r1 * jme + r2 * jme**2 + r3 * jme**3 + r4 * jme**4)/10**8
    return r


@jcompile('float64(float64)', nopython=True)
def geocentric_longitude(heliocentric_longitude):
    theta = heliocentric_longitude + 180.0
    return theta % 360


@jcompile('float64(float64)', nopython=True)
def geocentric_latitude(heliocentric_latitude):
    beta = -1.0*heliocentric_latitude
    return beta


@jcompile('float64(float64)', nopython=True)
def mean_elongation(julian_ephemeris_century):
    x0 = (297.85036
          + 445267.111480 * julian_ephemeris_century
          - 0.0019142 * julian_ephemeris_century**2
          + julian_ephemeris_century**3 / 189474)
    return x0


@jcompile('float64(float64)', nopython=True)
def mean_anomaly_sun(julian_ephemeris_century):
    x1 = (357.52772
          + 35999.050340 * julian_ephemeris_century
          - 0.0001603 * julian_ephemeris_century**2
          - julian_ephemeris_century**3 / 300000)
    return x1


@jcompile('float64(float64)', nopython=True)
def mean_anomaly_moon(julian_ephemeris_century):
    x2 = (134.96298
          + 477198.867398 * julian_ephemeris_century
          + 0.0086972 * julian_ephemeris_century**2
          + julian_ephemeris_century**3 / 56250)
    return x2


@jcompile('float64(float64)', nopython=True)
def moon_argument_latitude(julian_ephemeris_century):
    x3 = (93.27191
          + 483202.017538 * julian_ephemeris_century
          - 0.0036825 * julian_ephemeris_century**2
          + julian_ephemeris_century**3 / 327270)
    return x3


@jcompile('float64(float64)', nopython=True)
def moon_ascending_longitude(julian_ephemeris_century):
    x4 = (125.04452
          - 1934.136261 * julian_ephemeris_century
          + 0.0020708 * julian_ephemeris_century**2
          + julian_ephemeris_century**3 / 450000)
    return x4


@jcompile('float64(float64, float64, float64, float64, float64, float64)',
          nopython=True)
def longitude_nutation(julian_ephemeris_century, x0, x1, x2, x3, x4):
    delta_psi_sum = 0
    for row in range(NUTATION_YTERM_ARRAY.shape[0]):
        a = NUTATION_ABCD_ARRAY[row, 0]
        b = NUTATION_ABCD_ARRAY[row, 1]
        argsin = (NUTATION_YTERM_ARRAY[row, 0]*x0 +
                  NUTATION_YTERM_ARRAY[row, 1]*x1 +
                  NUTATION_YTERM_ARRAY[row, 2]*x2 +
                  NUTATION_YTERM_ARRAY[row, 3]*x3 +
                  NUTATION_YTERM_ARRAY[row, 4]*x4)
        term = (a + b * julian_ephemeris_century) * np.sin(np.radians(argsin))
        delta_psi_sum += term
    delta_psi = delta_psi_sum*1.0/36000000
    return delta_psi


@jcompile('float64(float64, float64, float64, float64, float64, float64)',
          nopython=True)
def obliquity_nutation(julian_ephemeris_century, x0, x1, x2, x3, x4):
    delta_eps_sum = 0.0
    for row in range(NUTATION_YTERM_ARRAY.shape[0]):
        c = NUTATION_ABCD_ARRAY[row, 2]
        d = NUTATION_ABCD_ARRAY[row, 3]
        argcos = (NUTATION_YTERM_ARRAY[row, 0]*x0 +
                  NUTATION_YTERM_ARRAY[row, 1]*x1 +
                  NUTATION_YTERM_ARRAY[row, 2]*x2 +
                  NUTATION_YTERM_ARRAY[row, 3]*x3 +
                  NUTATION_YTERM_ARRAY[row, 4]*x4)
        term = (c + d * julian_ephemeris_century) * np.cos(np.radians(argcos))
        delta_eps_sum += term
    delta_eps = delta_eps_sum*1.0/36000000
    return delta_eps


@jcompile('float64(float64)', nopython=True)
def mean_ecliptic_obliquity(julian_ephemeris_millennium):
    U = 1.0*julian_ephemeris_millennium/10
    e0 = (84381.448 - 4680.93 * U - 1.55 * U**2
          + 1999.25 * U**3 - 51.38 * U**4 - 249.67 * U**5
          - 39.05 * U**6 + 7.12 * U**7 + 27.87 * U**8
          + 5.79 * U**9 + 2.45 * U**10)
    return e0


@jcompile('float64(float64, float64)', nopython=True)
def true_ecliptic_obliquity(mean_ecliptic_obliquity, obliquity_nutation):
    e0 = mean_ecliptic_obliquity
    deleps = obliquity_nutation
    e = e0*1.0/3600 + deleps
    return e


@jcompile('float64(float64)', nopython=True)
def aberration_correction(earth_radius_vector):
    deltau = -20.4898 / (3600 * earth_radius_vector)
    return deltau


@jcompile('float64(float64, float64, float64)', nopython=True)
def apparent_sun_longitude(geocentric_longitude, longitude_nutation,
                           aberration_correction):
    lamd = geocentric_longitude + longitude_nutation + aberration_correction
    return lamd


@jcompile('float64(float64, float64)', nopython=True)
def mean_sidereal_time(julian_day, julian_century):
    v0 = (280.46061837 + 360.98564736629 * (julian_day - 2451545)
          + 0.000387933 * julian_century**2 - julian_century**3 / 38710000)
    return v0 % 360.0


@jcompile('float64(float64, float64, float64)', nopython=True)
def apparent_sidereal_time(mean_sidereal_time, longitude_nutation,
                           true_ecliptic_obliquity):
    v = mean_sidereal_time + longitude_nutation * np.cos(
        np.radians(true_ecliptic_obliquity))
    return v


@jcompile('float64(float64, float64, float64)', nopython=True)
def geocentric_sun_right_ascension(apparent_sun_longitude,
                                   true_ecliptic_obliquity,
                                   geocentric_latitude):
    num = (np.sin(np.radians(apparent_sun_longitude))
           * np.cos(np.radians(true_ecliptic_obliquity))
           - np.tan(np.radians(geocentric_latitude))
           * np.sin(np.radians(true_ecliptic_obliquity)))
    alpha = np.degrees(np.arctan2(num, np.cos(
        np.radians(apparent_sun_longitude))))
    return alpha % 360


@jcompile('float64(float64, float64, float64)', nopython=True)
def geocentric_sun_declination(apparent_sun_longitude, true_ecliptic_obliquity,
                               geocentric_latitude):
    delta = np.degrees(np.arcsin(np.sin(np.radians(geocentric_latitude)) *
                                 np.cos(np.radians(true_ecliptic_obliquity)) +
                                 np.cos(np.radians(geocentric_latitude)) *
                                 np.sin(np.radians(true_ecliptic_obliquity)) *
                                 np.sin(np.radians(apparent_sun_longitude))))
    return delta


@jcompile('float64(float64, float64, float64)', nopython=True)
def local_hour_angle(apparent_sidereal_time, observer_longitude,
                     sun_right_ascension):
    """Measured westward from south"""
    H = apparent_sidereal_time + observer_longitude - sun_right_ascension
    return H % 360


@jcompile('float64(float64)', nopython=True)
def equatorial_horizontal_parallax(earth_radius_vector):
    xi = 8.794 / (3600 * earth_radius_vector)
    return xi


@jcompile('float64(float64)', nopython=True)
def uterm(observer_latitude):
    u = np.arctan(0.99664719 * np.tan(np.radians(observer_latitude)))
    return u


@jcompile('float64(float64, float64, float64)', nopython=True)
def xterm(u, observer_latitude, observer_elevation):
    x = (np.cos(u) + observer_elevation / 6378140
         * np.cos(np.radians(observer_latitude)))
    return x


@jcompile('float64(float64, float64, float64)', nopython=True)
def yterm(u, observer_latitude, observer_elevation):
    y = (0.99664719 * np.sin(u) + observer_elevation / 6378140
         * np.sin(np.radians(observer_latitude)))
    return y


@jcompile('float64(float64, float64,float64, float64)', nopython=True)
def parallax_sun_right_ascension(xterm, equatorial_horizontal_parallax,
                                 local_hour_angle, geocentric_sun_declination):
    num = (-xterm * np.sin(np.radians(equatorial_horizontal_parallax))
           * np.sin(np.radians(local_hour_angle)))
    denom = (np.cos(np.radians(geocentric_sun_declination))
             - xterm * np.sin(np.radians(equatorial_horizontal_parallax))
             * np.cos(np.radians(local_hour_angle)))
    delta_alpha = np.degrees(np.arctan2(num, denom))
    return delta_alpha


@jcompile('float64(float64, float64)', nopython=True)
def topocentric_sun_right_ascension(geocentric_sun_right_ascension,
                                    parallax_sun_right_ascension):
    alpha_prime = geocentric_sun_right_ascension + parallax_sun_right_ascension
    return alpha_prime


@jcompile('float64(float64, float64, float64, float64, float64, float64)',
          nopython=True)
def topocentric_sun_declination(geocentric_sun_declination, xterm, yterm,
                                equatorial_horizontal_parallax,
                                parallax_sun_right_ascension,
                                local_hour_angle):
    num = ((np.sin(np.radians(geocentric_sun_declination)) - yterm
            * np.sin(np.radians(equatorial_horizontal_parallax)))
           * np.cos(np.radians(parallax_sun_right_ascension)))
    denom = (np.cos(np.radians(geocentric_sun_declination)) - xterm
             * np.sin(np.radians(equatorial_horizontal_parallax))
             * np.cos(np.radians(local_hour_angle)))
    delta = np.degrees(np.arctan2(num, denom))
    return delta


@jcompile('float64(float64, float64)', nopython=True)
def topocentric_local_hour_angle(local_hour_angle,
                                 parallax_sun_right_ascension):
    H_prime = local_hour_angle - parallax_sun_right_ascension
    return H_prime


@jcompile('float64(float64, float64, float64)', nopython=True)
def topocentric_elevation_angle_without_atmosphere(observer_latitude,
                                                   topocentric_sun_declination,
                                                   topocentric_local_hour_angle
                                                   ):
    e0 = np.degrees(np.arcsin(
        np.sin(np.radians(observer_latitude))
        * np.sin(np.radians(topocentric_sun_declination))
        + np.cos(np.radians(observer_latitude))
        * np.cos(np.radians(topocentric_sun_declination))
        * np.cos(np.radians(topocentric_local_hour_angle))))
    return e0


@jcompile('float64(float64, float64, float64, float64)', nopython=True)
def atmospheric_refraction_correction(local_pressure, local_temp,
                                      topocentric_elevation_angle_wo_atmosphere,
                                      atmos_refract):
    # switch sets delta_e when the sun is below the horizon
    switch = topocentric_elevation_angle_wo_atmosphere >= -1.0 * (
        0.26667 + atmos_refract)
    delta_e = ((local_pressure / 1010.0) * (283.0 / (273 + local_temp))
               * 1.02 / (60 * np.tan(np.radians(
                   topocentric_elevation_angle_wo_atmosphere
                   + 10.3 / (topocentric_elevation_angle_wo_atmosphere
                             + 5.11))))) * switch
    return delta_e


@jcompile('float64(float64, float64)', nopython=True)
def topocentric_elevation_angle(topocentric_elevation_angle_without_atmosphere,
                                atmospheric_refraction_correction):
    e = (topocentric_elevation_angle_without_atmosphere
         + atmospheric_refraction_correction)
    return e


@jcompile('float64(float64)', nopython=True)
def topocentric_zenith_angle(topocentric_elevation_angle):
    theta = 90 - topocentric_elevation_angle
    return theta


@jcompile('float64(float64, float64, float64)', nopython=True)
def topocentric_astronomers_azimuth(topocentric_local_hour_angle,
                                    topocentric_sun_declination,
                                    observer_latitude):
    num = np.sin(np.radians(topocentric_local_hour_angle))
    denom = (np.cos(np.radians(topocentric_local_hour_angle))
             * np.sin(np.radians(observer_latitude))
             - np.tan(np.radians(topocentric_sun_declination))
             * np.cos(np.radians(observer_latitude)))
    gamma = np.degrees(np.arctan2(num, denom))
    return gamma % 360


@jcompile('float64(float64)', nopython=True)
def topocentric_azimuth_angle(topocentric_astronomers_azimuth):
    phi = topocentric_astronomers_azimuth + 180
    return phi % 360


@jcompile('float64(float64)', nopython=True)
def sun_mean_longitude(julian_ephemeris_millennium):
    M = (280.4664567 + 360007.6982779 * julian_ephemeris_millennium
         + 0.03032028 * julian_ephemeris_millennium**2
         + julian_ephemeris_millennium**3 / 49931
         - julian_ephemeris_millennium**4 / 15300
         - julian_ephemeris_millennium**5 / 2000000)
    return M


@jcompile('float64(float64, float64, float64, float64)', nopython=True)
def equation_of_time(sun_mean_longitude, geocentric_sun_right_ascension,
                     longitude_nutation, true_ecliptic_obliquity):
    E = (sun_mean_longitude - 0.0057183 - geocentric_sun_right_ascension +
         longitude_nutation * np.cos(np.radians(true_ecliptic_obliquity)))
    # limit between 0 and 360
    E = E % 360
    # convert to minutes
    E *= 4
    greater = E > 20
    less = E < -20
    other = (E <= 20) & (E >= -20)
    E = greater * (E - 1440) + less * (E + 1440) + other * E
    return E


@jcompile('void(float64[:], float64[:], float64[:,:])', nopython=True,
          nogil=True)
def solar_position_loop(unixtime, loc_args, out):
    """Loop through the time array and calculate the solar position"""
    lat = loc_args[0]
    lon = loc_args[1]
    elev = loc_args[2]
    pressure = loc_args[3]
    temp = loc_args[4]
    delta_t = loc_args[5]
    atmos_refract = loc_args[6]
    sst = loc_args[7]
    esd = loc_args[8]

    for i in range(unixtime.shape[0]):
        utime = unixtime[i]
        jd = julian_day(utime)
        jde = julian_ephemeris_day(jd, delta_t)
        jc = julian_century(jd)
        jce = julian_ephemeris_century(jde)
        jme = julian_ephemeris_millennium(jce)
        R = heliocentric_radius_vector(jme)
        if esd:
            out[0, i] = R
            continue
        L = heliocentric_longitude(jme)
        B = heliocentric_latitude(jme)
        Theta = geocentric_longitude(L)
        beta = geocentric_latitude(B)
        x0 = mean_elongation(jce)
        x1 = mean_anomaly_sun(jce)
        x2 = mean_anomaly_moon(jce)
        x3 = moon_argument_latitude(jce)
        x4 = moon_ascending_longitude(jce)
        delta_psi = longitude_nutation(jce, x0, x1, x2, x3, x4)
        delta_epsilon = obliquity_nutation(jce, x0, x1, x2, x3, x4)
        epsilon0 = mean_ecliptic_obliquity(jme)
        epsilon = true_ecliptic_obliquity(epsilon0, delta_epsilon)
        delta_tau = aberration_correction(R)
        lamd = apparent_sun_longitude(Theta, delta_psi, delta_tau)
        v0 = mean_sidereal_time(jd, jc)
        v = apparent_sidereal_time(v0, delta_psi, epsilon)
        alpha = geocentric_sun_right_ascension(lamd, epsilon, beta)
        delta = geocentric_sun_declination(lamd, epsilon, beta)
        if sst:
            out[0, i] = v
            out[1, i] = alpha
            out[2, i] = delta
            continue
        m = sun_mean_longitude(jme)
        eot = equation_of_time(m, alpha, delta_psi, epsilon)
        H = local_hour_angle(v, lon, alpha)
        xi = equatorial_horizontal_parallax(R)
        u = uterm(lat)
        x = xterm(u, lat, elev)
        y = yterm(u, lat, elev)
        delta_alpha = parallax_sun_right_ascension(x, xi, H, delta)
        delta_prime = topocentric_sun_declination(delta, x, y, xi, delta_alpha,
                                                  H)
        H_prime = topocentric_local_hour_angle(H, delta_alpha)
        e0 = topocentric_elevation_angle_without_atmosphere(lat, delta_prime,
                                                            H_prime)
        delta_e = atmospheric_refraction_correction(pressure, temp, e0,
                                                    atmos_refract)
        e = topocentric_elevation_angle(e0, delta_e)
        theta = topocentric_zenith_angle(e)
        theta0 = topocentric_zenith_angle(e0)
        gamma = topocentric_astronomers_azimuth(H_prime, delta_prime, lat)
        phi = topocentric_azimuth_angle(gamma)
        out[0, i] = theta
        out[1, i] = theta0
        out[2, i] = e
        out[3, i] = e0
        out[4, i] = phi
        out[5, i] = eot


def solar_position_numba(unixtime, lat, lon, elev, pressure, temp, delta_t,
                         atmos_refract, numthreads, sst=False, esd=False):
    """Calculate the solar position using the numba compiled functions
    and multiple threads. Very slow if functions are not numba compiled.
    """
    # these args are the same for each thread
    loc_args = np.array([lat, lon, elev, pressure, temp, delta_t,
                         atmos_refract, sst, esd])

    # construct dims x ulength array to put the results in
    ulength = unixtime.shape[0]
    if sst:
        dims = 3
    elif esd:
        dims = 1
    else:
        dims = 6
    result = np.empty((dims, ulength), dtype=np.float64)

    if unixtime.dtype != np.float64:
        unixtime = unixtime.astype(np.float64)

    if ulength < numthreads:
        warnings.warn('The number of threads is more than the length of '
                      'the time array. Only using %s threads.'.format(ulength))
        numthreads = ulength

    if numthreads <= 1:
        solar_position_loop(unixtime, loc_args, result)
        return result

    # split the input and output arrays into numthreads chunks
    split0 = np.array_split(unixtime, numthreads)
    split2 = np.array_split(result, numthreads, axis=1)
    chunks = [[a0, loc_args, split2[i]] for i, a0 in enumerate(split0)]
    # Spawn one thread per chunk
    threads = [threading.Thread(target=solar_position_loop, args=chunk)
               for chunk in chunks]
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()
    return result


def solar_position_numpy(unixtime, lat, lon, elev, pressure, temp, delta_t,
                         atmos_refract, numthreads, sst=False, esd=False):
    """Calculate the solar position assuming unixtime is a numpy array. Note
    this function will not work if the solar position functions were
    compiled with numba.
    """

    jd = julian_day(unixtime)
    jde = julian_ephemeris_day(jd, delta_t)
    jc = julian_century(jd)
    jce = julian_ephemeris_century(jde)
    jme = julian_ephemeris_millennium(jce)
    R = heliocentric_radius_vector(jme)
    if esd:
        return (R, )
    L = heliocentric_longitude(jme)
    B = heliocentric_latitude(jme)
    Theta = geocentric_longitude(L)
    beta = geocentric_latitude(B)
    x0 = mean_elongation(jce)
    x1 = mean_anomaly_sun(jce)
    x2 = mean_anomaly_moon(jce)
    x3 = moon_argument_latitude(jce)
    x4 = moon_ascending_longitude(jce)
    delta_psi = longitude_nutation(jce, x0, x1, x2, x3, x4)
    delta_epsilon = obliquity_nutation(jce, x0, x1, x2, x3, x4)
    epsilon0 = mean_ecliptic_obliquity(jme)
    epsilon = true_ecliptic_obliquity(epsilon0, delta_epsilon)
    delta_tau = aberration_correction(R)
    lamd = apparent_sun_longitude(Theta, delta_psi, delta_tau)
    v0 = mean_sidereal_time(jd, jc)
    v = apparent_sidereal_time(v0, delta_psi, epsilon)
    alpha = geocentric_sun_right_ascension(lamd, epsilon, beta)
    delta = geocentric_sun_declination(lamd, epsilon, beta)
    if sst:
        return v, alpha, delta
    m = sun_mean_longitude(jme)
    eot = equation_of_time(m, alpha, delta_psi, epsilon)
    H = local_hour_angle(v, lon, alpha)
    xi = equatorial_horizontal_parallax(R)
    u = uterm(lat)
    x = xterm(u, lat, elev)
    y = yterm(u, lat, elev)
    delta_alpha = parallax_sun_right_ascension(x, xi, H, delta)
    delta_prime = topocentric_sun_declination(delta, x, y, xi, delta_alpha, H)
    H_prime = topocentric_local_hour_angle(H, delta_alpha)
    e0 = topocentric_elevation_angle_without_atmosphere(lat, delta_prime,
                                                        H_prime)
    delta_e = atmospheric_refraction_correction(pressure, temp, e0,
                                                atmos_refract)
    e = topocentric_elevation_angle(e0, delta_e)
    theta = topocentric_zenith_angle(e)
    theta0 = topocentric_zenith_angle(e0)
    gamma = topocentric_astronomers_azimuth(H_prime, delta_prime, lat)
    phi = topocentric_azimuth_angle(gamma)
    return theta, theta0, e, e0, phi, eot


def solar_position(unixtime, lat, lon, elev, pressure, temp, delta_t,
                   atmos_refract, numthreads=8, sst=False, esd=False):

    """
    Calculate the solar position using the
    NREL SPA algorithm described in [1].

    If numba is installed, the functions can be compiled
    and the code runs quickly. If not, the functions
    still evaluate but use numpy instead.

    Parameters
    ----------
    unixtime : numpy array
        Array of unix/epoch timestamps to calculate solar position for.
        Unixtime is the number of seconds since Jan. 1, 1970 00:00:00 UTC.
        A pandas.DatetimeIndex is easily converted using .astype(np.int64)/10**9
    lat : float
        Latitude to calculate solar position for
    lon : float
        Longitude to calculate solar position for
    elev : float
        Elevation of location in meters
    pressure : int or float
        avg. yearly pressure at location in millibars;
        used for atmospheric correction
    temp : int or float
        avg. yearly temperature at location in
        degrees C; used for atmospheric correction
    delta_t : float
        Difference between terrestrial time and UT1.
    atmos_refrac : float
        The approximate atmospheric refraction (in degrees)
        at sunrise and sunset.
    numthreads: int, optional, default 8
        Number of threads to use for computation if numba>=0.17
        is installed.
    sst : bool, default False
        If True, return only data needed for sunrise, sunset, and transit
        calculations.
    esd : bool, default False
        If True, return only Earth-Sun distance in AU

    Returns
    -------
    Numpy Array with elements:
        apparent zenith,
        zenith,
        elevation,
        apparent_elevation,
        azimuth,
        equation_of_time

    References
    ----------
    [1] I. Reda and A. Andreas, Solar position algorithm for solar radiation
    applications. Solar Energy, vol. 76, no. 5, pp. 577-589, 2004.

    [2] I. Reda and A. Andreas, Corrigendum to Solar position algorithm for
    solar radiation applications. Solar Energy, vol. 81, no. 6, p. 838, 2007.
    """
    if USE_NUMBA:
        do_calc = solar_position_numba
    else:
        do_calc = solar_position_numpy

    result = do_calc(unixtime, lat, lon, elev, pressure,
                     temp, delta_t, atmos_refract, numthreads,
                     sst, esd)

    if not isinstance(result, np.ndarray):
        try:
            result = np.array(result)
        except Exception:
            pass

    return result


def transit_sunrise_sunset(dates, lat, lon, delta_t, numthreads):
    """
    Calculate the sun transit, sunrise, and sunset
    for a set of dates at a given location.

    Parameters
    ----------
    dates : array
        Numpy array of ints/floats corresponding to the Unix time
        for the dates of interest, must be midnight UTC (00:00+00:00)
        on the day of interest.
    lat : float
        Latitude of location to perform calculation for
    lon : float
        Longitude of location
    delta_t : float
        Difference between terrestrial time and UT. USNO has tables.
    numthreads : int
        Number to threads to use for calculation (if using numba)

    Returns
    -------
    tuple : (transit, sunrise, sunset) localized to UTC

    """

    if ((dates % 86400) != 0.0).any():
        raise ValueError('Input dates must be at 00:00 UTC')

    utday = (dates // 86400) * 86400
    ttday0 = utday - delta_t
    ttdayn1 = ttday0 - 86400
    ttdayp1 = ttday0 + 86400

    # index 0 is v, 1 is alpha, 2 is delta
    utday_res = solar_position(utday, 0, 0, 0, 0, 0, delta_t,
                               0, numthreads, sst=True)
    v = utday_res[0]

    ttday0_res = solar_position(ttday0, 0, 0, 0, 0, 0, delta_t,
                                0, numthreads, sst=True)
    ttdayn1_res = solar_position(ttdayn1, 0, 0, 0, 0, 0, delta_t,
                                 0, numthreads, sst=True)
    ttdayp1_res = solar_position(ttdayp1, 0, 0, 0, 0, 0, delta_t,
                                 0, numthreads, sst=True)
    m0 = (ttday0_res[1] - lon - v) / 360
    cos_arg = ((np.sin(np.radians(-0.8333)) - np.sin(np.radians(lat))
               * np.sin(np.radians(ttday0_res[2]))) /
               (np.cos(np.radians(lat)) * np.cos(np.radians(ttday0_res[2]))))
    cos_arg[abs(cos_arg) > 1] = np.nan
    H0 = np.degrees(np.arccos(cos_arg)) % 180

    m = np.empty((3, len(utday)))
    m[0] = m0 % 1
    m[1] = (m[0] - H0 / 360)
    m[2] = (m[0] + H0 / 360)

    # need to account for fractions of day that may be the next or previous
    # day in UTC
    add_a_day = m[2] >= 1
    sub_a_day = m[1] < 0
    m[1] = m[1] % 1
    m[2] = m[2] % 1
    vs = v + 360.985647 * m
    n = m + delta_t / 86400

    a = ttday0_res[1] - ttdayn1_res[1]
    a[abs(a) > 2] = a[abs(a) > 2] % 1
    ap = ttday0_res[2] - ttdayn1_res[2]
    ap[abs(ap) > 2] = ap[abs(ap) > 2] % 1
    b = ttdayp1_res[1] - ttday0_res[1]
    b[abs(b) > 2] = b[abs(b) > 2] % 1
    bp = ttdayp1_res[2] - ttday0_res[2]
    bp[abs(bp) > 2] = bp[abs(bp) > 2] % 1
    c = b - a
    cp = bp - ap

    alpha_prime = ttday0_res[1] + (n * (a + b + c * n)) / 2
    delta_prime = ttday0_res[2] + (n * (ap + bp + cp * n)) / 2
    Hp = (vs + lon - alpha_prime) % 360
    Hp[Hp >= 180] = Hp[Hp >= 180] - 360

    h = np.degrees(np.arcsin(np.sin(np.radians(lat)) *
                             np.sin(np.radians(delta_prime)) +
                             np.cos(np.radians(lat)) *
                             np.cos(np.radians(delta_prime))
                             * np.cos(np.radians(Hp))))

    T = (m[0] - Hp[0] / 360) * 86400
    R = (m[1] + (h[1] + 0.8333) / (360 * np.cos(np.radians(delta_prime[1])) *
                                   np.cos(np.radians(lat)) *
                                   np.sin(np.radians(Hp[1])))) * 86400
    S = (m[2] + (h[2] + 0.8333) / (360 * np.cos(np.radians(delta_prime[2])) *
                                   np.cos(np.radians(lat)) *
                                   np.sin(np.radians(Hp[2])))) * 86400

    S[add_a_day] += 86400
    R[sub_a_day] -= 86400

    transit = T + utday
    sunrise = R + utday
    sunset = S + utday

    return transit, sunrise, sunset


def earthsun_distance(unixtime, delta_t, numthreads):
    """
    Calculates the distance from the earth to the sun using the
    NREL SPA algorithm described in [1].

    Parameters
    ----------
    unixtime : numpy array
        Array of unix/epoch timestamps to calculate solar position for.
        Unixtime is the number of seconds since Jan. 1, 1970 00:00:00 UTC.
        A pandas.DatetimeIndex is easily converted using .astype(np.int64)/10**9
    delta_t : float
        Difference between terrestrial time and UT. USNO has tables.
    numthreads : int
        Number to threads to use for calculation (if using numba)

    Returns
    -------
    R : array
        Earth-Sun distance in AU.

    References
    ----------
    [1] Reda, I., Andreas, A., 2003. Solar position algorithm for solar
    radiation applications. Technical report: NREL/TP-560- 34302. Golden,
    USA, http://www.nrel.gov.
    """

    R = solar_position(unixtime, 0, 0, 0, 0, 0, delta_t,
                       0, numthreads, esd=True)[0]

    return R


def calculate_deltat(year, month):
    """Calculate the difference between Terrestrial Dynamical Time (TD)
    and Universal Time (UT).

    Note: This function is not yet compatible for calculations using
    Numba.

    Equations taken from http://eclipse.gsfc.nasa.gov/SEcat5/deltatpoly.html
    """

    plw = 'Deltat is unknown for years before -1999 and after 3000. ' \
          'Delta values will be calculated, but the calculations ' \
          'are not intended to be used for these years.'

    try:
        if np.any((year > 3000) | (year < -1999)):
            warnings.warn(plw)
    except ValueError:
        if (year > 3000) | (year < -1999):
            warnings.warn(plw)
    except TypeError:
        return 0

    y = year + (month - 0.5)/12

    deltat = np.where(year < -500,

                      -20+32*((y-1820)/100)**2, 0)

    deltat = np.where((-500 <= year) & (year < 500),

                      10583.6-1014.41*(y/100)
                      + 33.78311*(y/100)**2
                      - 5.952053*(y/100)**3
                      - 0.1798452*(y/100)**4
                      + 0.022174192*(y/100)**5
                      + 0.0090316521*(y/100)**6, deltat)

    deltat = np.where((500 <= year) & (year < 1600),

                      1574.2-556.01*((y-1000)/100)
                      + 71.23472*((y-1000)/100)**2
                      + 0.319781*((y-1000)/100)**3
                      - 0.8503463*((y-1000)/100)**4
                      - 0.005050998*((y-1000)/100)**5
                      + 0.0083572073*((y-1000)/100)**6, deltat)

    deltat = np.where((1600 <= year) & (year < 1700),

                      120-0.9808*(y-1600)
                      - 0.01532*(y-1600)**2
                      + (y-1600)**3/7129, deltat)

    deltat = np.where((1700 <= year) & (year < 1800),

                      8.83+0.1603*(y-1700)
                      - 0.0059285*(y-1700)**2
                      + 0.00013336*(y-1700)**3
                      - (y-1700)**4/1174000, deltat)

    deltat = np.where((1800 <= year) & (year < 1860),

                      13.72-0.332447*(y-1800)
                      + 0.0068612*(y-1800)**2
                      + 0.0041116*(y-1800)**3
                      - 0.00037436*(y-1800)**4
                      + 0.0000121272*(y-1800)**5
                      - 0.0000001699*(y-1800)**6
                      + 0.000000000875*(y-1800)**7, deltat)

    deltat = np.where((1860 <= year) & (year < 1900),

                      7.62+0.5737*(y-1860)
                      - 0.251754*(y-1860)**2
                      + 0.01680668*(y-1860)**3
                      - 0.0004473624*(y-1860)**4
                      + (y-1860)**5/233174, deltat)

    deltat = np.where((1900 <= year) & (year < 1920),

                      -2.79+1.494119*(y-1900)
                      - 0.0598939*(y-1900)**2
                      + 0.0061966*(y-1900)**3
                      - 0.000197*(y-1900)**4, deltat)

    deltat = np.where((1920 <= year) & (year < 1941),

                      21.20+0.84493*(y-1920)
                      - 0.076100*(y-1920)**2
                      + 0.0020936*(y-1920)**3, deltat)

    deltat = np.where((1941 <= year) & (year < 1961),

                      29.07+0.407*(y-1950)
                      - (y-1950)**2/233
                      + (y-1950)**3/2547, deltat)

    deltat = np.where((1961 <= year) & (year < 1986),

                      45.45+1.067*(y-1975)
                      - (y-1975)**2/260
                      - (y-1975)**3/718, deltat)

    deltat = np.where((1986 <= year) & (year < 2005),

                      63.86+0.3345*(y-2000)
                      - 0.060374*(y-2000)**2
                      + 0.0017275*(y-2000)**3
                      + 0.000651814*(y-2000)**4
                      + 0.00002373599*(y-2000)**5, deltat)

    deltat = np.where((2005 <= year) & (year < 2050),

                      62.92+0.32217*(y-2000)
                      + 0.005589*(y-2000)**2, deltat)

    deltat = np.where((2050 <= year) & (year < 2150),

                      -20+32*((y-1820)/100)**2
                      - 0.5628*(2150-y), deltat)

    deltat = np.where(year >= 2150,

                      -20+32*((y-1820)/100)**2, deltat)

    deltat = deltat.item() if np.isscalar(year) & np.isscalar(month)\
        else deltat

    return deltat
